from typing import Union, Literal
from datetime import datetime

import pandas as pd

from econuy.utils.transform import error_handler


def _convert_usd(
    data: pd.DataFrame,
    metadata: "Metadata",  # type: ignore # noqa: F821
    error_handling: Literal["raise", "coerce", "ignore"] = "raise",
) -> pd.DataFrame:
    from econuy.load import load_dataset

    indicators = metadata.indicator_ids
    metadata = metadata.copy()
    # We get the first one because we validated that all indicators have the same metadata, or pass them one by one
    single_metadata = metadata.indicator_metadata[indicators[0]]

    if single_metadata["currency"] != "UYU":
        output = error_handler(data, errors=error_handling, msg="Currency is not UYU")
        return output, metadata

    nxr = load_dataset("nxr_monthly")
    time_series_type = single_metadata["time_series_type"]
    inferred_freq = pd.infer_freq(data.index)
    target_freq = inferred_freq

    # For now we only support converting monthly or lower frequency data, so we resample first.
    if inferred_freq in ["D", "B", "C", "W", "W-SUN", None]:
        if time_series_type == "Flow":
            data = data.resample("ME").sum()
        else:
            data = data.resample("ME").last()
        target_freq = "ME"

    if time_series_type == "Stock":
        nxr_freq = nxr.resample(target_freq, operation="last").data.iloc[:, [1]]
    else:
        cum_periods = single_metadata["cumulative_periods"]
        nxr_freq = (
            nxr.resample(target_freq, operation="mean")
            .rolling(window=cum_periods, operation="mean")
            .data.iloc[:, [0]]
        )

    nxr_to_use = nxr_freq.reindex(data.index).iloc[:, 0]
    output = data.div(nxr_to_use, axis=0)
    metadata.update_dataset_metadata({"currency": "USD"})
    metadata.add_transformation_step({"convert": {"flavor": "usd"}})

    return output, metadata


def _convert_real(
    data: pd.DataFrame,
    metadata: "Metadata",  # type: ignore # noqa: F821
    start_date: Union[str, datetime, None] = None,
    end_date: Union[str, datetime, None] = None,
    error_handling: Literal["raise", "coerce", "ignore"] = "raise",
) -> pd.DataFrame:
    from econuy.load import load_dataset

    indicators = metadata.indicator_ids
    metadata = metadata.copy()
    # We get the first one because we validated that all indicators have the same metadata, or pass them one by one
    single_metadata = metadata.indicator_metadata[indicators[0]]

    if single_metadata["currency"] != "UYU":
        output = error_handler(data, errors=error_handling, msg="Currency is not UYU")
        return output, metadata
    elif "Constant" in str(single_metadata["inflation_adjustment"]):
        output = error_handler(
            data, errors=error_handling, msg="Already inflation adjusted"
        )
        return output, metadata

    cpi = load_dataset("cpi")

    inferred_freq = pd.infer_freq(data.index)
    target_freq = inferred_freq
    if inferred_freq in ["D", "B", "C", "W", "W-SUN", None]:
        if single_metadata["time_series_type"] == "Flow":
            data = data.resample("ME").sum()
        else:
            data = data.resample("ME").mean()
        target_freq = pd.infer_freq(data.index)

    cum_periods = single_metadata["cumulative_periods"]
    cpi_to_use = (
        cpi.resample(target_freq, operation="mean")
        .rolling(cum_periods, operation="mean")
        .data.iloc[:, 0]
    )

    start_date = (
        datetime.strptime(start_date, "%Y-%m-%d")
        if isinstance(start_date, str)
        else start_date
    )
    end_date = (
        datetime.strptime(end_date, "%Y-%m-%d")
        if isinstance(end_date, str)
        else end_date
    )

    if start_date is None:
        converted_df = data.div(cpi_to_use, axis=0)
        col_text = "Constant prices"
    elif end_date is None:
        # start_date = pd.to_datetime(start_date)
        month = data.index.to_series().sub(pd.to_datetime(start_date)).abs().idxmin()
        converted_df = data.div(cpi_to_use, axis=0) * cpi_to_use.loc[month]
        m_start = start_date.strftime("%Y-%m")
        col_text = f"Constant prices {m_start}"
    else:
        converted_df = (
            data.div(cpi_to_use, axis=0) * cpi_to_use[start_date:end_date].mean()
        )
        m_start = start_date.strftime("%Y-%m")
        m_end = end_date.strftime("%Y-%m")
        if m_start == m_end:
            col_text = f"Constant prices {m_start}"
        else:
            col_text = f"Constant prices {m_start}_{m_end}"

    converted_df = converted_df.reindex(data.index)
    metadata.update_dataset_metadata({"inflation_adjustment": col_text})
    start_date_str = start_date.strftime("%Y-%m-%d") if start_date is not None else None
    end_date_str = end_date.strftime("%Y-%m-%d") if end_date is not None else None
    metadata.add_transformation_step(
        {
            "convert": {
                "flavor": "real",
                "start_date": start_date_str,
                "end_date": end_date_str,
            }
        }
    )

    return converted_df, metadata


def _convert_gdp(
    data: pd.DataFrame,
    metadata: "Metadata",  # type: ignore # noqa: F821
    error_handling: Literal["raise", "coerce", "ignore"] = "raise",
) -> pd.DataFrame:
    from econuy.load import load_dataset

    indicators = metadata.indicator_ids
    metadata = metadata.copy()
    # We get the first one because we validated that all indicators have the same metadata, or pass them one by one
    single_metadata = metadata.indicator_metadata[indicators[0]]

    if single_metadata["currency"] not in ["UYU", "USD"]:
        output = error_handler(
            data, errors=error_handling, msg="Currency is not UYU or USD"
        )
        return output, metadata
    elif single_metadata["unit"] == "% GDP":
        output = error_handler(data, errors=error_handling, msg="Already in % GDP")
        return output, metadata
    elif single_metadata["inflation_adjustment"] is not None:
        output = error_handler(
            data, errors=error_handling, msg="Data is inflation adjusted"
        )
        return output, metadata

    gdp = load_dataset("gdp_denominator").data

    inferred_freq = pd.infer_freq(data.index)
    target_freq = inferred_freq
    cum_periods = single_metadata["cumulative_periods"]
    ts_type = single_metadata["time_series_type"]

    if inferred_freq in ["M", "MS", "ME"]:
        gdp = gdp.resample(target_freq).interpolate("linear")
        if cum_periods != 12 and ts_type == "Flow":
            converter = int(12 / cum_periods)
            data = data.rolling(window=converter).sum()
    elif inferred_freq in ["Q", "QE-DEC", "QE-DEC"]:
        gdp = gdp.resample(inferred_freq).asfreq()
        if cum_periods != 4 and ts_type == "Flow":
            converter = int(4 / cum_periods)
            data = data.rolling(window=converter).sum()
    elif inferred_freq in ["A", "A-DEC", "YE-DEC"]:
        gdp = gdp.resample(inferred_freq).asfreq()
    elif inferred_freq in ["D", "B", "C", "W", "W-SUN", None]:
        if ts_type == "Flow":
            data = data.resample("ME").sum()
        else:
            data = data.resample("ME").mean()
        gdp = gdp.resample("ME").interpolate("linear")
    else:
        raise ValueError(
            "Frequency of input dataframe not any of 'D', 'C', "
            "'W', 'B', 'M', 'MS', 'Q', 'QE-DEC', 'A' or 'A-DEC'."
        )

    currency = single_metadata["currency"]
    if currency == "USD":
        gdp = gdp.iloc[:, 1].to_frame()
    else:
        gdp = gdp.iloc[:, 0].to_frame()

    gdp_to_use = gdp.reindex(data.index).iloc[:, 0]
    converted_df = data.div(gdp_to_use, axis=0).multiply(100)

    metadata.update_dataset_metadata({"unit": "% GDP"})
    metadata.add_transformation_step({"convert": {"flavor": "gdp"}})

    return converted_df, metadata



import warnings
from typing import Union, Tuple

import pandas as pd
import numpy as np


def _resample(
    data: pd.DataFrame,
    metadata: "Metadata",  # type: ignore # noqa: F821
    rule: Union[pd.DateOffset, pd.Timedelta, str],
    operation: str = "sum",
    interpolation: str = "linear",
) -> Tuple[pd.DataFrame, "Metadata"]:  # type: ignore # noqa: F821
    pd_frequencies = {
        "A": 1,
        "A-DEC": 1,
        "YE-DEC": 1,
        "Q": 4,
        "Q-DEC": 4,
        "QE-DEC": 4,
        "M": 12,
        "ME": 12,
        "W": 52.143,
        "W-SUN": 52.143,
        "2W": 26.071,
        "2W-SUN": 26.071,
        "B": 240,
        "D": 365,
    }
    indicators = metadata.indicator_ids
    metadata = metadata.copy()
    # We get the first one because we validated that all indicators have the same metadata, or pass them one by one
    single_metadata = metadata.indicator_metadata[indicators[0]]

    if operation == "sum":
        output = data.resample(rule).sum()
    elif operation == "mean":
        output = data.resample(rule).mean()
    elif operation == "last":
        output = data.resample(rule).last()
    else:
        output = data.resample(rule).last()
        output = output.interpolate(method=interpolation)

    cum_periods = single_metadata["cumulative_periods"]

    if cum_periods != 1:
        input_notna = data.iloc[:, 0].count()
        output_notna = output.iloc[:, 0].count()
        cum_adj = round(output_notna / input_notna)
        metadata.update_dataset_metadata({"cumulative_periods": cum_adj})

    if operation in ["sum", "mean", "last"]:
        infer_base = pd.infer_freq(data.index)
        try:
            base_freq = pd_frequencies[infer_base]
            target_freq = pd_frequencies[rule]
            if target_freq < base_freq:
                count = int(base_freq / target_freq)
                proc = data.resample(rule).count()
                antimask = np.where(proc >= count, False, True)
                output = output.mask(antimask, np.nan)
        except KeyError:
            warnings.warn(
                "No bin trimming performed because frequencies "
                "could not be assigned a numeric value",
                UserWarning,
            )
    metadata.add_transformation_step(
        {
            "resample": {
                "rule": rule,
                "operation": operation,
                "interpolation": interpolation,
            }
        }
    )
    output = output.dropna(how="all")

    return output, metadata



import pandas as pd

from econuy.base import Dataset, DatasetMetadata
from econuy.utils.operations import get_name_from_function, get_download_sources


def income_household() -> Dataset:
    """Get average household income.

    Returns
    -------
    Monthly average household income : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = (
        pd.read_excel(sources["main"], skiprows=5)
        .dropna(thresh=5)
        .loc[lambda x: x["Mes, Trimestre y Año"].str.contains("/[0-9]{2}", regex=True)]
    )

    output = raw.set_index(
        pd.date_range(start="2006-03-31", freq="QE-DEC", periods=len(raw))
    ).iloc[:, 1:]

    spanish_names = [
        "Total país",
        "Montevideo",
        "Interior: total",
        "Interior: localidades de más de 5 mil hab.",
        "Interior: localidades pequeñas y rural",
    ]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]
    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    base_metadata = {
        "area": "Income and expenditure",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Pesos",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def income_capita() -> Dataset:
    """Get average per capita income.

    Returns
    -------
    Monthly average per capita income : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = (
        pd.read_excel(sources["main"], skiprows=5)
        .dropna(thresh=5)
        .loc[lambda x: x["Mes, Trimestre y Año "].str.contains("/[0-9]{2}", regex=True)]
    )

    output = raw.set_index(
        pd.date_range(start="2006-03-31", freq="QE-DEC", periods=len(raw))
    ).iloc[:, 1:]

    output.columns = [
        "Total país",
        "Montevideo",
        "Interior: total",
        "Interior: localidades de más de 5 mil hab.",
        "Interior: localidades pequeñas y rural",
    ]

    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Income and expenditure",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Pesos",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset



import datetime as dt
import time
from io import BytesIO, StringIO
from urllib.error import HTTPError, URLError
from httpx import ConnectError

import pandas as pd
import httpx
from pandas.tseries.offsets import MonthEnd

from econuy.utils.chromedriver import _build
from econuy.utils.operations import get_download_sources, get_name_from_function
from econuy.utils.retrieval import get_with_ssl_context
from econuy.base import Dataset, DatasetMetadata


def bank_credit() -> pd.DataFrame:
    """Get bank credit data.

    Returns
    -------
    Monthly credit : pd.DataFrame

    """
    name = get_name_from_function()
    sources = get_download_sources(name)
    try:
        xls = pd.ExcelFile(sources["main"])
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            r_bytes = get_with_ssl_context("bcu", sources["main"])
            xls = pd.ExcelFile(r_bytes)
    tc = pd.read_excel(
        xls,
        sheet_name="TC",
        skiprows=1,
        usecols="A:B",
        index_col=0,
        parse_dates=True,
        date_format="%Y%m",
    ).squeeze()
    tc.index = tc.index + MonthEnd(0)
    raw = pd.read_excel(
        xls,
        sheet_name="Total Sist. Banc.",
        skiprows=10,
        usecols="A:P,T:AB,AD:AL",
        index_col=0,
    )
    raw.index = pd.to_datetime(raw.index, errors="coerce")
    output = raw.loc[~pd.isna(raw.index)].dropna(how="all", axis=1)
    output.index = output.index + MonthEnd(0)
    output.columns = [
        "Créditos: Resid. privado - vigentes",
        "Créditos: Resid. privado - vencidos",
        "Créditos: Resid. privado- total",
        "Créditos: Resid. público - vigentes",
        "Créditos: Resid. público - vencidos",
        "Créditos: Resid. público - total",
        "Créditos: Resid. total - vigentes",
        "Créditos: Resid. total - vencidos",
        "Créditos: Resid. total - total",
        "Créditos: No residentes - vigentes",
        "Créditos: No residentes - vencidos",
        "Créditos: No residentes - total",
        "Créditos: Total - vigentes",
        "Créditos: Total - vencidos",
        "Créditos: Total - total",
        "Créditos: Resid. MN privado - vigentes",
        "Créditos: Resid. MN privado - vencidos",
        "Créditos: Resid. MN privado- total",
        "Créditos: Resid. MN público - vigentes",
        "Créditos: Resid. MN público - vencidos",
        "Créditos: Resid. MN público- total",
        "Créditos: Resid. MN total - vigentes",
        "Créditos: Resid. MN total - vencidos",
        "Créditos: Resid. MN total- total",
        "Créditos: Resid. ME privado - vigentes",
        "Créditos: Resid. ME privado - vencidos",
        "Créditos: Resid. ME privado- total",
        "Créditos: Resid. ME público - vigentes",
        "Créditos: Resid. ME público - vencidos",
        "Créditos: Resid. ME público- total",
        "Créditos: Resid. ME total - vigentes",
        "Créditos: Resid. ME total - vencidos",
        "Créditos: Resid. ME total- total",
    ]
    output.iloc[:, :24] = output.iloc[:, :24].div(tc, axis=0)

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    spanish_names = output.columns
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Financial market",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def bank_deposits() -> pd.DataFrame:
    """Get bank deposits data.

    Returns
    -------
    Monthly deposits : pd.DataFrame

    """
    name = get_name_from_function()
    sources = get_download_sources(name)
    try:
        xls = pd.ExcelFile(sources["main"])
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            r_bytes = get_with_ssl_context("bcu", sources["main"])
            xls = pd.ExcelFile(r_bytes)
    tc = pd.read_excel(
        xls,
        sheet_name="TC",
        skiprows=1,
        usecols="A:B",
        index_col=0,
        parse_dates=True,
        date_format="%Y%m",
    ).squeeze()
    tc.index = tc.index + MonthEnd(0)
    raw = pd.read_excel(
        xls,
        sheet_name="Total Sist. Banc.",
        skiprows=8,
        usecols="A:S",
        index_col=0,
    )
    raw.index = pd.to_datetime(raw.index, errors="coerce")
    output = raw.loc[~pd.isna(raw.index)].dropna(how="all", axis=1)
    output.index = output.index + MonthEnd(0)
    output.columns = [
        "Depósitos: S. privado - MN",
        "Depósitos: S. privado - ME",
        "Depósitos: S. privado - total",
        "Depósitos: S. público - MN",
        "Depósitos: S. público - ME",
        "Depósitos: S. público - total",
        "Depósitos: Total - MN",
        "Depósitos: Total - ME",
        "Depósitos: Total - total",
        "Depósitos: S. privado - MN vista",
        "Depósitos: S. privado - MN plazo",
        "Depósitos: S. privado - ME vista",
        "Depósitos: S. privado - ME plazo",
        "Depósitos: S. privado - total vista",
        "Depósitos: S. privado - total plazo",
        "Depósitos: S. privado - residente",
        "Depósitos: S. privado - no residente",
    ]
    output.iloc[:, :15] = output.iloc[:, :15].div(tc, axis=0)
    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    spanish_names = output.columns
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Financial market",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def bank_interest_rates() -> pd.DataFrame:
    """Get interest rates data.

    Returns
    -------
    Monthly interest rates : pd.DataFrame

    """
    name = get_name_from_function()
    sources = get_download_sources(name)
    try:
        xls = pd.ExcelFile(sources["main"])
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            r_bytes = get_with_ssl_context("bcu", sources["main"])
            xls = pd.ExcelFile(r_bytes)

    sheets = [
        "Activas $",
        "Activas UI",
        "Activas U$S",
        "Pasivas $",
        "Pasivas UI",
        "Pasivas U$S",
    ]
    columns = ["B:C,G,K", "B:C,G,K", "B:C,H,L", "B:C,N,T", "B:C,P,W", "B:C,N,T"]
    sheet_data = []
    for sheet, columns in zip(sheets, columns):
        if "Activas" in sheet:
            skip = 11
        else:
            skip = 10
        data = pd.read_excel(
            xls, sheet_name=sheet, skiprows=skip, usecols=columns, index_col=0
        )
        data.index = pd.to_datetime(data.index, errors="coerce")
        data = data.loc[~pd.isna(data.index)]
        data.index = data.index + MonthEnd(0)
        sheet_data.append(data)
    output = pd.concat(sheet_data, axis=1)
    output.columns = [
        "Tasas activas: $, promedio",
        "Tasas activas: $, promedio empresas",
        "Tasas activas: $, promedio familias",
        "Tasas activas: UI, promedio",
        "Tasas activas: UI, promedio empresas",
        "Tasas activas: UI, promedio familias",
        "Tasas activas: US$, promedio",
        "Tasas activas: US$, promedio empresas",
        "Tasas activas: US$, promedio familias",
        "Tasas pasivas: $, promedio",
        "Tasas pasivas: $, promedio empresas",
        "Tasas pasivas: $, promedio familias",
        "Tasas pasivas: UI, promedio",
        "Tasas pasivas: UI, promedio empresas",
        "Tasas pasivas: UI, promedio familias",
        "Tasas pasivas: US$, promedio",
        "Tasas pasivas: US$, promedio empresas",
        "Tasas pasivas: US$, promedio familias",
    ]

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    spanish_names = output.columns
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Financial market",
        "currency": "-",
        "inflation_adjustment": "-",
        "unit": "Rate",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    for indicator, currency in zip(
        ids,
        [
            "UYU",
            "UYU",
            "UYU",
            "UYU",
            "UYU",
            "UYU",
            "USD",
            "USD",
            "USD",
            "UYU",
            "UYU",
            "UYU",
            "UYU",
            "UYU",
            "UYU",
            "USD",
            "USD",
            "USD",
        ],
    ):
        metadata.update_indicator_metadata_value(indicator, "currency", currency)
    for indicator, const in zip(
        ids,
        [
            None,
            None,
            None,
            "Const.",
            "Const.",
            "Const.",
            None,
            None,
            None,
            None,
            None,
            None,
            "Const.",
            "Const.",
            "Const.",
            None,
            None,
            None,
        ],
    ):
        metadata.update_indicator_metadata_value(
            indicator, "inflation_adjustment", const
        )
    dataset = Dataset(name, output, metadata)

    return dataset


def sovereign_risk_index() -> pd.DataFrame:
    """Get Uruguayan Bond Index (sovereign risk spreads) data.

    Returns
    -------
    Uruguayan Bond Index : pd.DataFrame

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    try:
        historical = pd.read_excel(
            sources["historical"],
            usecols="B:C",
            skiprows=1,
            index_col=0,
            sheet_name="Valores de Cierre Diarios",
        )
    except (ConnectError, URLError, HTTPError):
        r_historical = httpx.get(sources["historical"], verify=False)
        historical = pd.read_excel(
            BytesIO(r_historical.content),
            usecols="B:C",
            skiprows=1,
            index_col=0,
            sheet_name="Valores de Cierre Diarios",
        )
    driver = _build()
    driver.get(sources["current"])
    text = driver.page_source
    current = (
        pd.read_html(StringIO(text))[0]
        .set_index("Fecha")
        .rename_axis(None)
        .rename(columns={"Valor": "UBI"})
    )
    current.index = pd.to_datetime(current.index, format="%d/%m/%y")
    output = pd.concat([historical, current])
    output = output.loc[~output.index.duplicated(keep="last")]
    output.sort_index(inplace=True)
    output = output.apply(pd.to_numeric, errors="coerce").interpolate(
        limit_area="inside"
    )
    output = output.rename_axis(None)

    spanish_names = output.columns
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Financial market",
        "currency": "USD",
        "inflation_adjustment": None,
        "unit": "Bps",
        "seasonal_adjustment": None,
        "frequency": "D",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def _bypass_bevsa_disclaimer_maybe(driver, url: str):
    driver.get(url)
    if "Disclaimer.aspx" in driver.current_url:
        checkbox = driver.find_element(
            by="id", value="ContentPlaceHolder1_chkAcceptTerms"
        )
        checkbox.click()
        time.sleep(2)

        accept_button = driver.find_element(
            by="id", value="ContentPlaceHolder1_btnContinue"
        )
        accept_button.click()
        time.sleep(2)

        driver.get(url)


def call_rate() -> pd.DataFrame:
    """Get 1-day call interest rate data.

    Returns
    -------
    Daily call rate : pd.DataFrame

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    driver = _build()
    _bypass_bevsa_disclaimer_maybe(driver, sources["main"])

    start = driver.find_element(
        by="name", value="ctl00$ContentPlaceHolder1$dateDesde$dateInput"
    )
    start.clear()
    start.send_keys("01/01/2002")
    end = driver.find_element(
        by="name", value="ctl00$ContentPlaceHolder1$dateHasta$dateInput"
    )
    end.clear()
    end.send_keys(dt.datetime.now().strftime("%d/%m/%Y"))
    submit = driver.find_element(by="id", value="ContentPlaceHolder1_LinkFiltrar")
    submit.click()
    time.sleep(5)
    tables = pd.read_html(StringIO(driver.page_source), decimal=",", thousands=".")
    driver.quit()

    raw = tables[8].iloc[:, :-2]
    output = raw.set_index("FECHA")
    output.index = pd.to_datetime(output.index, format="%d/%m/%Y")
    output.columns = [
        "Tasa call a 1 día: Promedio",
        "Tasa call a 1 día: Máximo",
        "Tasa call a 1 día: Mínimo",
    ]
    output = output.sort_index()
    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    spanish_names = output.columns
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Financial market",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Rate",
        "seasonal_adjustment": None,
        "frequency": "D",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def sovereign_bond_yields() -> pd.DataFrame:
    """Get interest rate yield for Uruguayan US-denominated bonds,
    inflation-linked bonds and peso bonds.

    Returns
    -------
    Daily bond yields in basis points : pd.DataFrame

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    driver = _build()
    _bypass_bevsa_disclaimer_maybe(driver, sources["usd"])
    dfs = []
    for url in sources.values():
        driver.get(url)
        start = driver.find_element(
            by="name", value="ctl00$ContentPlaceHolder1$dateDesde$dateInput"
        )
        start.clear()
        start.send_keys("01/01/2000")
        end = driver.find_element(
            by="name", value="ctl00$ContentPlaceHolder1$dateHasta$dateInput"
        )
        end.clear()
        end.send_keys(dt.datetime.now().strftime("%d/%m/%Y"))
        submit = driver.find_element(by="id", value="ContentPlaceHolder1_LinkFiltrar")
        submit.click()
        time.sleep(5)
        tables = pd.read_html(StringIO(driver.page_source), decimal=",", thousands=".")

        raw = tables[8]
        df = raw.set_index("FECHA")
        df.index = pd.to_datetime(df.index, format="%d/%m/%Y")
        df.sort_index(inplace=True)
        df = df.loc[:, df.columns.isin(["BPS", "RENDIMIENTO"])]
        df.columns = [url]
        dfs.append(df)
    driver.quit()
    output = dfs[0].join(dfs[1], how="outer").join(dfs[2], how="outer")
    output.columns = [
        "Bonos soberanos en dólares",
        "Bonos soberanos en UI",
        "Bonos soberanos en pesos",
    ]
    output = output.loc[~output.index.duplicated()]

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    spanish_names = output.columns
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Financial market",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Rate",
        "seasonal_adjustment": None,
        "frequency": "D",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    for indicator, const in zip(ids, [None, "Const.", None]):
        metadata.update_indicator_metadata_value(
            indicator, "inflation_adjustment", const
        )
    for indicator, currency in zip(ids, ["USD", "UYU", "UYU"]):
        metadata.update_indicator_metadata_value(indicator, "currency", currency)
    dataset = Dataset(name, output, metadata)

    return dataset



from os import path

from sqlalchemy import create_engine

from econuy.utils.metadata import _get_sources
from econuy.utils import sql

try:
    from tests.test_session import remove_temporary_files_folders
except ImportError:
    from .test_session import remove_temporary_files_folders


CUR_DIR = path.abspath(path.dirname(__file__))
TEST_DIR = path.join(CUR_DIR, "test-data")
TEST_CON = create_engine("sqlite://").connect()
sql.insert_csvs(con=TEST_CON, directory=TEST_DIR)


def test_sqlutil():
    remove_temporary_files_folders()
    sql.read(con=TEST_CON, command="SELECT * FROM nxr_daily")
    sql.read(
        con=TEST_CON,
        table_name="nxr_daily",
        start_date="2011-01-14",
        end_date="2012-01-15",
    )
    sql.read(con=TEST_CON, table_name="nxr_daily", start_date="2011-01-14")
    sql.read(
        con=TEST_CON,
        table_name="international_reserves_changes",
        end_date="2012-01-01",
        cols=["1. Compras netas de moneda extranjera"],
    )


def test_sources():
    source_1 = _get_sources("fiscal_balance_global_public_sector")
    source_2 = _get_sources("fiscal_balance_nonfinancial_public_sector")
    assert source_1 == source_2



import warnings
from typing import Union, Dict, Tuple, Literal, Optional

import pandas as pd
from statsmodels.tools.sm_exceptions import X13Error, X13Warning
from statsmodels.tsa import x13 as x13_sm
from statsmodels.tsa.x13 import x13_arima_analysis as x13a
from statsmodels.tsa.seasonal import STL, seasonal_decompose, MSTL

from econuy.utils import x13 as x13_utils
from econuy.utils.transform import error_handler


# The `_open_and_read` function needs to be monkey-patched to specify the
# encoding or decomposition will fail on Windows
def _new_open_and_read(fname):
    with open(fname, "r", encoding="utf8") as fin:
        fout = fin.read()
    return fout


x13_sm._open_and_read = _new_open_and_read


def _decompose(
    data: pd.DataFrame,
    metadata: "Metadata",  # type: ignore # noqa: F821
    component: Literal["t-c", "sa"] = "sa",
    method: Literal["x13", "loess", "mloess", "moving_averages"] = "x13",
    fallback: str = "loess",
    fn_kwargs: Optional[Dict] = None,
    ignore_warnings: bool = True,
    error_handling: Literal["raise", "coerce", "ignore"] = "raise",
) -> Union[Tuple[pd.DataFrame, pd.DataFrame], pd.DataFrame]:
    indicators = metadata.indicator_ids
    metadata = metadata.copy()
    # We get the first one because we validated that all indicators have the same metadata, or pass them one by one
    single_metadata = metadata.indicator_metadata[indicators[0]]

    if single_metadata["seasonal_adjustment"] is not None:
        output = error_handler(
            data,
            errors=error_handling,
            msg="Data has already been seasonally adjusted.",
        )
        return output, metadata

    data_proc = data.copy()
    columns = data_proc.columns
    trends_array = []
    seas_adjs_array = []
    for col in data_proc.columns:
        col_df = data_proc[col].dropna()
        if method == "x13":
            x13_binary_path = x13_utils._get_binary()
            try:
                with warnings.catch_warnings():
                    if ignore_warnings is True:
                        action = "ignore"
                    else:
                        action = "default"
                    warnings.filterwarnings(action=action, category=X13Warning)
                    results = x13a(
                        col_df,
                        x12path=x13_binary_path,
                        prefer_x13=True,
                        **fn_kwargs,
                    )

                    trends = results.trend.reindex(data_proc.index).T
                    seas_adjs = results.seasadj.reindex(data_proc.index).T

            except X13Error:
                print(f"X13 error. Falling back to {fallback}")
                if fallback == "loess":
                    results = STL(col_df).fit()
                elif fallback == "mloess":
                    results = MSTL(col_df).fit()
                else:
                    results = seasonal_decompose(col_df, extrapolate_trend="freq")
                trends = results.trend.reindex(data_proc.index).T
                seas_adjs = (
                    (results.observed - results.seasonal).reindex(data_proc.index).T
                )

        else:
            if method == "loess":
                results = STL(col_df, **fn_kwargs).fit()
            elif method == "mloess":
                results = MSTL(col_df, **fn_kwargs).fit()
            else:
                results = seasonal_decompose(
                    col_df, extrapolate_trend="freq", **fn_kwargs
                )

            trends = results.trend.reindex(data_proc.index).T
            seas_adjs = (results.observed - results.seasonal).reindex(data_proc.index).T

        trends_array.append(trends)
        seas_adjs_array.append(seas_adjs)
    trends = pd.concat(trends_array, axis=1)
    seas_adjs = pd.concat(seas_adjs_array, axis=1)
    trends.columns = columns
    seas_adjs.columns = columns

    if component == "sa":
        output = seas_adjs
        metadata_value = "Seasonally adjusted"
    elif component == "t-c":
        output = trends
        metadata_value = "Trend-cycle"

    metadata.update_dataset_metadata({"seasonal_adjustment": metadata_value})
    metadata.add_transformation_step(
        {"decompose": {"component": component, "method": method}}
    )

    return output, metadata



import platform
import os
from pathlib import Path

import httpx


def _get_binary() -> None:
    if "X13PATH" in os.environ:
        return os.environ["X13PATH"]

    urls = {
        "Windows": "https://raw.githubusercontent.com/rxavier/econuy-extras/main/econuy_extras/x13/windows/x13as.exe",
        "Darwin-x64": "https://raw.githubusercontent.com/rxavier/econuy-extras/main/econuy_extras/x13/darwin/x64/x13as",
        "Darwin-arm64": "https://raw.githubusercontent.com/rxavier/econuy-extras/main/econuy_extras/x13/darwin/arm64/x13as",
        "Linux-x64": "https://raw.githubusercontent.com/rxavier/econuy-extras/main/econuy_extras/x13/linux/x64/x13as",
        "Linux-arm64": "https://raw.githubusercontent.com/rxavier/econuy-extras/main/econuy_extras/x13/linux/arm64/x13as",
    }

    system_string = platform.system()
    if system_string == "Windows":
        suffix = ".exe"
        base_dir = Path.home() / "AppData" / "Roaming" / "econuy"
    else:
        suffix = ""
        base_dir = Path.home() / ".econuy"
    if system_string == "Darwin":
        if "ARM64" in platform.version():
            system_string += "-arm64"
        else:
            system_string += "-x64"
        base_dir = Path.home() / "Library" / "Application Support" / "econuy"
    elif system_string == "Linux":
        if "aarch64" in platform.machine():
            system_string += "-arm64"
        else:
            system_string += "-x64"
    if system_string not in urls.keys():
        raise ValueError(
            "X13 binaries are only available for Windows, Darwin (macOS) or Linux."
        )

    base_dir.mkdir(exist_ok=True, parents=True)
    binary_path = Path(base_dir, f"x13as{suffix}")

    if binary_path.exists():
        os.chmod(binary_path, 0o755)
        os.environ["X13PATH"] = binary_path.as_posix()
        print(f"Using existing binary at {binary_path}")
        return binary_path.as_posix()

    r = httpx.get(urls[system_string])
    r.raise_for_status()
    with open(binary_path, "wb") as f:
        f.write(r.content)
    os.chmod(binary_path, 0o755)
    os.environ["X13PATH"] = binary_path.as_posix()

    print(
        f"Download complete. Saved binary to {base_dir} and set X13PATH to {binary_path}"
    )

    return binary_path.as_posix()



import os

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service


def _build(download_dir: str = "."):
    chrome_options = Options()
    chrome_options.add_argument("--headless")
    chrome_options.add_argument("--disable-gpu")
    chrome_options.add_argument("--no-sandbox")
    chrome_options.add_experimental_option(
        "prefs",
        {
            "download.default_directory": download_dir,
            "download.prompt_for_download": False,
            "download.directory_upgrade": True,
        },
    )

    if os.environ.get("IN_DOCKER") == "true":  # fix for webapp
        service = Service(executable_path="/usr/bin/chromedriver")
        chrome_options.add_argument("--disable-dev-shm-usage")
    else:
        service = None

    return webdriver.Chrome(service=service, options=chrome_options)



import datetime as dt
import re
import tempfile
import time
from pathlib import Path
from os import listdir, path

import pandas as pd
import numpy as np
import patoolib
import httpx
from pandas.tseries.offsets import MonthEnd
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

from econuy.utils.operations import get_download_sources, get_name_from_function
from econuy.utils.chromedriver import _build
from econuy.base import Dataset, DatasetMetadata
from econuy import load_dataset
from econuy.utils.retrieval import get_with_ssl_context


def monthly_gdp() -> Dataset:
    """Get the monthly indicator for economic activity.

    Returns
    -------
    Monthly GDP : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    with tempfile.TemporaryDirectory() as tmp_dir:
        driver = _build(tmp_dir)
        driver.get(sources["main"])
        button = WebDriverWait(driver, 3).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, "button.dt-button.buttons-excel.buttons-html5")
            )
        )
        driver.execute_script("arguments[0].click()", button)
        time.sleep(3)
        driver.quit()
        filepath = list(Path(tmp_dir).glob("*.xlsx"))[0]
        output = pd.read_excel(filepath, index_col=0)

    output.index = pd.date_range(start="2016-01-31", freq="ME", periods=len(output))
    output = output.rename_axis(None)

    spanish_names = [
        "Indicador mensual de actividad económica",
        "Indicador mensual de actividad económica (desestacionalizado)",
        "Indicador mensual de actividad económica (tendencia-ciclo)",
    ]
    spanish_names = [{"es": x} for x in spanish_names]

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": "Constant prices 2016",
        "unit": "2016=100",
        "seasonal_adjustment": "",
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }

    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    for indicator, adjustment in zip(ids, [None, "Seasonally adjusted", "Trend-cycle"]):
        metadata = metadata.update_indicator_metadata_value(
            indicator, "seasonal_adjustment", adjustment
        )
    dataset = Dataset(name, output, metadata)

    return dataset


def national_accounts_supply_constant_nsa() -> Dataset:
    """Get supply-side national accounts data in NSA constant prices, 2005-.

    Returns
    -------
    National accounts, supply side, constant prices, NSA : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    r_bytes = get_with_ssl_context("bcu", sources["main"])
    raw = pd.read_excel(r_bytes, skiprows=7)
    output = raw.dropna(how="all", axis=1).iloc[:, 2:].dropna(how="all").T
    output.index = pd.date_range(start="2016-03-31", freq="QE-DEC", periods=len(output))
    output = output.apply(pd.to_numeric, errors="coerce").rename_axis(None)

    spanish_names = [
        "Agropecuario, pesca y minería",
        "Industrias manufactureras",
        "Energía eléctrica, gas y agua",
        "Construcción",
        "Comercio, alojamiento y suministro de comidas y bebidas",
        "Transporte y almacenamiento, información y comunicaciones",
        "Servicios financieros",
        "Actividades profesionales y arrendamiento",
        "Actividades de administración pública",
        "Salud, educación, act. inmobiliarias y otros servicios",
        "Valor agregado a precios básicos",
        "Impuestos menos subvenciones",
        "Producto bruto interno",
    ]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": "Constant prices 2016",
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def national_accounts_demand_constant_nsa() -> Dataset:
    """Get demand-side national accounts data in NSA constant prices, 2005-.

    Returns
    -------
    National accounts, demand side, constant prices, NSA : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    r_bytes = get_with_ssl_context("bcu", sources["main"])
    raw = pd.read_excel(r_bytes, skiprows=7)
    output = raw.dropna(how="all", axis=1).iloc[:, 2:].dropna(how="all").T
    output.index = pd.date_range(start="2016-03-31", freq="QE-DEC", periods=len(output))
    output = output.apply(pd.to_numeric, errors="coerce").rename_axis(None)
    output[7] = output[7] * -1

    spanish_names = [
        "Gasto de consumo: total",
        "Gasto de consumo: hogares",
        "Gasto de consumo: gobierno y ISFLH",
        "Formación bruta de capital",
        "Formación bruta de capital: fijo",
        "Variación de existencias",
        "Exportaciones de bienes y servicios",
        "Importaciones de bienes y servicios",
        "Producto bruto interno",
    ]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": "Constant prices 2016",
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def national_accounts_demand_current_nsa() -> Dataset:
    """Get demand-side national accounts data in NSA current prices.

    Returns
    -------
    National accounts, demand side, current prices, NSA : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    r_bytes = get_with_ssl_context("bcu", sources["main"])
    raw = pd.read_excel(r_bytes, skiprows=7)
    output = raw.dropna(how="all", axis=1).iloc[:, 2:].dropna(how="all").T
    output.index = pd.date_range(start="2016-03-31", freq="QE-DEC", periods=len(output))
    output = output.apply(pd.to_numeric, errors="coerce").rename_axis(None)
    output[7] = output[7] * -1

    spanish_names = [
        "Gasto de consumo: total",
        "Gasto de consumo: hogares",
        "Gasto de consumo: gobierno y ISFLH",
        "Formación bruta de capital",
        "Formación bruta de capital: fijo",
        "Variación de existencias",
        "Exportaciones de bienes y servicios",
        "Importaciones de bienes y servicios",
        "Producto bruto interno",
    ]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def national_accounts_supply_current_nsa() -> Dataset:
    """Get supply-side national accounts data in NSA current prices, 2005-.

    Returns
    -------
    National accounts, supply side, current prices, NSA : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    r_bytes = get_with_ssl_context("bcu", sources["main"])
    raw = pd.read_excel(r_bytes, skiprows=7)
    output = raw.dropna(how="all", axis=1).iloc[:, 2:].dropna(how="all").T
    output.index = pd.date_range(start="2016-03-31", freq="QE-DEC", periods=len(output))
    output = output.apply(pd.to_numeric, errors="coerce").rename_axis(None)

    spanish_names = [
        "Agropecuario, pesca y minería",
        "Industrias manufactureras",
        "Energía eléctrica, gas y agua",
        "Construcción",
        "Comercio, alojamiento y suministro de comidas y bebidas",
        "Transporte y almacenamiento, información y comunicaciones",
        "Servicios financieros",
        "Actividades profesionales y arrendamiento",
        "Actividades de administración pública",
        "Salud, educación, act. inmobiliarias y otros servicios",
        "Valor agregado a precios básicos",
        "Impuestos menos subvenciones",
        "Producto bruto interno",
    ]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def gdp_index_constant_sa() -> Dataset:
    """Get supply-side national accounts data in SA real index, 1997-.

    Returns
    -------
    National accounts, supply side, real index, SA : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    r_bytes = get_with_ssl_context("bcu", sources["main"])
    raw = pd.read_excel(r_bytes, skiprows=7)
    output = raw.dropna(how="all", axis=1).iloc[:, 2:].dropna(how="all").T
    output.index = pd.date_range(start="2016-03-31", freq="QE-DEC", periods=len(output))
    output = output.apply(pd.to_numeric, errors="coerce").rename_axis(None)

    spanish_names = ["Producto bruto interno"]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": "Constant prices 2016",
        "unit": "2016=100",
        "seasonal_adjustment": "Seasonally adjusted",
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def national_accounts_supply_constant_nsa_extended(
    *args,
    **kwargs,
) -> Dataset:
    """Get supply-side national accounts data in NSA constant prices, 1988-.

    Three datasets with different base years, 1983, 2005 and 2016, are spliced
    in order to get to the result DataFrame.

    Returns
    -------
    National accounts, supply side, constant prices, NSA : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    data_16 = load_dataset(
        "national_accounts_supply_constant_nsa", *args, **kwargs
    ).to_detailed()
    data_16.columns = data_16.columns.get_level_values(0)
    data_16["Otros servicios"] = (
        data_16["Servicios financieros"]
        + data_16["Actividades profesionales y arrendamiento"]
        + data_16["Actividades de administración pública"]
        + data_16["Salud, educación, act. inmobiliarias y otros servicios"]
    )
    data_16 = data_16.drop(
        [
            "Valor agregado a precios básicos",
            "Servicios financieros",
            "Actividades profesionales y arrendamiento",
            "Actividades de administración pública",
            "Salud, educación, act. inmobiliarias y otros servicios",
            "Valor agregado a precios básicos",
        ],
        axis=1,
    )

    names_05 = [
        "Actividades primarias",
        "Act. prim.: Agricultura, ganadería, caza y silvicultura",
        "Industrias manufactureras",
        "Suministro de electricidad, gas y agua",
        "Construcción",
        "Comercio, reparaciones, restaurantes y hoteles",
        "Transporte, almacenamiento y comunicaciones",
        "Otras actividades",
        "Otras actividades: SIFMI",
        "Impuestos menos subvenciones",
        "Producto bruto interno",
    ]
    r_bytes = get_with_ssl_context("bcu", sources["2005"])
    raw_05 = pd.read_excel(r_bytes, skiprows=9)
    data_05 = (
        raw_05.dropna(how="all", axis=1)
        .iloc[:, 1:]
        .set_index("Unnamed: 1")
        .dropna(how="all")
        .T
    )
    data_05.index = pd.date_range(
        start="1997-03-31", freq="QE-DEC", periods=len(data_05)
    )
    data_05 = data_05.apply(pd.to_numeric, errors="coerce").rename_axis(None)
    data_05.columns = names_05

    # In the following lines I rename and reorder the Other activities
    # column, and explicitly drop SIFMI because I have no way of adding them.
    # In order for this to be correct, I would have to substract the SIFMI
    # part from each of the other sectors, which I cannot do. Implicitly I'm
    # assuming that the share of SIFMI in each sector over time doesn't change,
    # which shouldn't affect growth rates.
    data_05["Otros servicios"] = data_05["Otras actividades"]
    data_05 = data_05.drop(
        [
            "Act. prim.: Agricultura, ganadería, caza y silvicultura",
            "Otras actividades",
            "Otras actividades: SIFMI",
        ],
        axis=1,
    )
    data_05.columns = data_16.columns
    aux_index = list(dict.fromkeys(list(data_05.index) + list(data_16.index)))
    aux = data_16.reindex(aux_index)
    for quarter in reversed(aux_index):
        if aux.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            aux.loc[quarter, :] = (
                aux.loc[next_quarter, :]
                * data_05.loc[quarter, :]
                / data_05.loc[next_quarter, :]
            )
    spanish_names = [
        "Agropecuario, pesca y minería",
        "Industrias manufactureras",
        "Energía eléctrica, gas y agua",
        "Construcción",
        "Comercio, alojamiento y suministro de comidas y bebidas",
        "Transporte y almacenamiento, información y comunicaciones",
        "Otros servicios",
        "Impuestos menos subvenciones",
        "Producto bruto interno",
    ]
    aux = aux[spanish_names]
    r_bytes = get_with_ssl_context("bcu", sources["1983"])
    data_83 = (
        pd.read_excel(
            r_bytes,
            skiprows=10,
            nrows=8,
            index_col=1,
        )
        .iloc[:, 1:]
        .T
    )
    data_83.index = pd.date_range(
        start="1988-03-31", freq="QE-DEC", periods=len(data_83)
    )
    data_83["Impuestos menos subvenciones"] = np.nan
    data_83 = data_83[
        [
            "AGROPECUARIA",
            "INDUSTRIAS MANUFACTURERAS",
            "SUMINISTRO DE ELECTRICIDAD, GAS Y AGUA",
            "CONSTRUCCION",
            "COMERCIO, REPARACIONES, RESTAURANTES Y HOTELES",
            "TRANSPORTE, ALMACENAMIENTO Y COMUNICACIONES",
            "OTRAS (1)",
            "Impuestos menos subvenciones",
            "PRODUCTO INTERNO BRUTO",
        ]
    ]
    data_83.columns = aux.columns

    complete_index = list(dict.fromkeys(list(data_83.index) + list(aux.index)))
    output = aux.reindex(complete_index)
    for quarter in reversed(complete_index):
        if output.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            output.loc[quarter, :] = (
                output.loc[next_quarter, :]
                * data_83.loc[quarter, :]
                / data_83.loc[next_quarter, :]
            )

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": "Constant prices 2016",
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def national_accounts_demand_constant_nsa_extended(*args, **kwargs) -> Dataset:
    """Get demand-side national accounts data in NSA constant prices, 1988-.

    Three datasets with different base years, 1983, 2005 and 2016, are spliced
    in order to get to the result DataFrame.

    Returns
    -------
    National accounts, demand side, constant prices, NSA : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    data_16 = load_dataset(
        "national_accounts_demand_constant_nsa", *args, **kwargs
    ).to_detailed()
    data_16.columns = data_16.columns.get_level_values(0)
    data_16 = data_16.drop(["Variación de existencias"], axis=1)

    data_05_names = [
        "Gasto de consumo final",
        "Gasto de consumo final de hogares",
        "Gasto de consumo final del gobierno general",
        "Formación bruta de capital",
        "Formación bruta de capital fijo",
        "Sector público",
        "Sector privado",
        "Exportaciones",
        "Importaciones",
        "Producto bruto interno",
    ]
    r_bytes = get_with_ssl_context("bcu", sources["2005"])
    raw_05 = pd.read_excel(r_bytes, skiprows=9)
    data_05 = (
        raw_05.dropna(how="all", axis=1)
        .iloc[:, 1:]
        .set_index("Unnamed: 1")
        .dropna(how="all")
        .T
    )
    data_05.index = pd.date_range(
        start="2005-03-31", freq="QE-DEC", periods=len(data_05)
    )
    data_05 = data_05.apply(pd.to_numeric, errors="coerce").rename_axis(None)
    data_05.columns = data_05_names
    data_05["Importaciones"] = data_05["Importaciones"] * -1

    data_05 = data_05.drop(["Sector público", "Sector privado"], axis=1)
    data_05.columns = data_16.columns

    aux_index = list(dict.fromkeys(list(data_05.index) + list(data_16.index)))
    aux = data_16.reindex(aux_index)
    for quarter in reversed(aux_index):
        if aux.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            aux.loc[quarter, :] = (
                aux.loc[next_quarter, :]
                * data_05.loc[quarter, :]
                / data_05.loc[next_quarter, :]
            )

    r_bytes = get_with_ssl_context("bcu", sources["1983"])
    data_83 = (
        pd.read_excel(
            r_bytes,
            skiprows=10,
            nrows=11,
            index_col=1,
        )
        .iloc[:, 1:]
        .T
    )
    data_83.index = pd.date_range(
        start="1988-03-31", freq="QE-DEC", periods=len(data_83)
    )
    data_83.drop(
        ["Sector público", "Sector privado", "Variación de existencias"],
        axis=1,
        inplace=True,
    )
    data_83.columns = aux.columns

    complete_index = list(dict.fromkeys(list(data_83.index) + list(aux.index)))
    output = aux.reindex(complete_index)
    for quarter in reversed(complete_index):
        if output.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            output.loc[quarter, :] = (
                output.loc[next_quarter, :]
                * data_83.loc[quarter, :]
                / data_83.loc[next_quarter, :]
            )
    spanish_names = [
        "Gasto de consumo final",
        "Gasto de consumo final de hogares",
        "Gasto de consumo final del gobierno general",
        "Formación bruta de capital",
        "Formación bruta de capital fijo",
        "Exportaciones",
        "Importaciones",
        "Producto bruto interno",
    ]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": "Constant prices 2016",
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def gdp_index_constant_sa_extended(*args, **kwargs) -> Dataset:
    """Get GDP data in SA constant prices, 1988-.

    Three datasets with different base years, 1983, 2005 and 2016, are spliced
    in order to get to the result DataFrame.

    Returns
    -------
    GDP, constant prices, SA : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    data_16 = load_dataset("gdp_index_constant_sa", *args, **kwargs).to_detailed()
    data_16.columns = data_16.columns.get_level_values(0)

    names = [
        "Actividades primarias",
        "Act. prim.: Agricultura, ganadería, caza y silvicultura",
        "Industrias manufactureras",
        "Suministro de electricidad, gas y agua",
        "Construcción",
        "Comercio, reparaciones, restaurantes y hoteles",
        "Transporte, almacenamiento y comunicaciones",
        "Otras actividades",
        "Otras actividades: SIFMI",
        "Impuestos menos subvenciones",
        "Producto bruto interno",
    ]
    r_bytes = get_with_ssl_context("bcu", sources["2005"])
    raw_05 = pd.read_excel(r_bytes, skiprows=9)
    data_05 = (
        raw_05.dropna(how="all", axis=1)
        .iloc[:, 1:]
        .set_index("Unnamed: 1")
        .dropna(how="all")
        .T
    )
    data_05.index = pd.date_range(
        start="1997-03-31", freq="QE-DEC", periods=len(data_05)
    )
    data_05 = data_05.apply(pd.to_numeric, errors="coerce").rename_axis(None)
    data_05.columns = names

    data_05 = data_05[["Producto bruto interno"]]
    aux_index = list(dict.fromkeys(list(data_05.index) + list(data_16.index)))
    aux = data_16.reindex(aux_index)
    for quarter in reversed(aux_index):
        if aux.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            aux.loc[quarter, :] = (
                aux.loc[next_quarter, :]
                * data_05.loc[quarter, :]
                / data_05.loc[next_quarter, :]
            )

    r_bytes = get_with_ssl_context("bcu", sources["1983"])
    data_83 = (
        pd.read_excel(
            r_bytes,
            skiprows=10,
            nrows=8,
            index_col=1,
        )
        .iloc[:, 1:]
        .T
    )
    data_83.index = pd.date_range(
        start="1988-03-31", freq="QE-DEC", periods=len(data_83)
    )
    data_83 = data_83[["PRODUCTO INTERNO BRUTO"]]
    data_83.columns = aux.columns

    complete_index = list(dict.fromkeys(list(data_83.index) + list(aux.index)))
    output = aux.reindex(complete_index)
    for quarter in reversed(complete_index):
        if output.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            output.loc[quarter, :] = (
                output.loc[next_quarter, :]
                * data_83.loc[quarter, :]
                / data_83.loc[next_quarter, :]
            )

    spanish_names = ["Producto bruto interno"]

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": "Constant prices 2016",
        "unit": "2016=100",
        "seasonal_adjustment": "Seasonally adjusted",
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def gdp_constant_nsa_extended(*args, **kwargs) -> Dataset:
    """Get GDP data in NSA constant prices, 1988-.

    Three datasets with two different base years, 1983 and 2016, are
    spliced in order to get to the result DataFrame. It uses the BCU's working
    paper for retropolated GDP in current and constant prices for 1997-2015.

    Returns
    -------
    GDP, constant prices, NSA : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    data_16 = load_dataset(
        "national_accounts_supply_constant_nsa", *args, **kwargs
    ).to_detailed()
    data_16.columns = data_16.columns.get_level_values(0)
    data_16 = data_16[["Producto bruto interno"]]

    names = ["Producto bruto interno"]

    r_bytes = get_with_ssl_context("bcu", sources["1997"])
    raw_97 = pd.read_excel(r_bytes, skiprows=6)
    data_97 = (
        raw_97.dropna(how="all", axis=1)
        .iloc[:, 1:]
        .set_index("Unnamed: 1")
        .dropna(how="all")
        .T
    )
    data_97.index = pd.date_range(
        start="1997-03-31", freq="QE-DEC", periods=len(data_97)
    )
    data_97 = data_97.apply(pd.to_numeric, errors="coerce").rename_axis(None)
    data_97.columns = names

    aux = pd.concat([data_97, data_16], axis=0)

    r_bytes = get_with_ssl_context("bcu", sources["1983"])
    data_83 = (
        pd.read_excel(
            r_bytes,
            skiprows=10,
            nrows=8,
            index_col=1,
        )
        .iloc[:, 1:]
        .T
    )
    data_83.index = pd.date_range(
        start="1988-03-31", freq="QE-DEC", periods=len(data_83)
    )
    data_83 = data_83[["PRODUCTO INTERNO BRUTO"]]
    data_83.columns = ["Producto bruto interno"]

    reaux_index = list(dict.fromkeys(list(data_83.index) + list(aux.index)))
    output = aux.reindex(reaux_index)
    for quarter in reversed(reaux_index):
        if output.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            output.loc[quarter, :] = (
                output.loc[next_quarter, :]
                * data_83.loc[quarter, :]
                / data_83.loc[next_quarter, :]
            )

    spanish_names = ["Producto bruto interno"]

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": "Constant prices 2016",
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def gdp_current_nsa_extended(*args, **kwargs) -> Dataset:
    """Get GDP data in NSA current prices, 1997-.

    It uses the BCU's working paper for retropolated GDP in current and constant prices for
    1997-2015.

    Returns
    -------
    GDP, current prices, NSA : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    data_16 = load_dataset(
        "national_accounts_supply_current_nsa", *args, **kwargs
    ).to_detailed()
    data_16.columns = data_16.columns.get_level_values(0)
    data_16 = data_16[["Producto bruto interno"]]

    names = ["Producto bruto interno"]

    r_bytes = get_with_ssl_context("bcu", sources["1997"])
    raw_97 = pd.read_excel(r_bytes, skiprows=6)
    data_97 = (
        raw_97.dropna(how="all", axis=1)
        .iloc[:, 1:]
        .set_index("Unnamed: 1")
        .dropna(how="all")
        .T
    )
    data_97.index = pd.date_range(
        start="1997-03-31", freq="QE-DEC", periods=len(data_97)
    )
    data_97 = data_97.apply(pd.to_numeric, errors="coerce").rename_axis(None)
    data_97.columns = names

    output = pd.concat([data_97, data_16], axis=0)

    spanish_names = ["Producto bruto interno"]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def gdp_denominator(*args, **kwargs):
    """Get nominal GDP data in UYU and USD with forecasts.

    Update nominal GDP data for use in the `transform.convert_gdp()` function.
    Get IMF forecasts for year of last available data point and the next
    year (for example, if the last period available at the BCU website is
    september 2019, fetch forecasts for 2019 and 2020).

    Returns
    -------
    output : Pandas dataframe
        Quarterly GDP in UYU and USD with 1 year forecasts.

    """
    name = get_name_from_function()

    data_uyu = load_dataset("gdp_current_nsa_extended", *args, **kwargs).rolling(
        window=4, operation="sum"
    )
    data_usd = data_uyu.convert("usd")
    data_uyu, data_usd = data_uyu.data, data_usd.data

    data = [data_uyu, data_usd]
    last_year = data_uyu.index.max().year
    if data_uyu.index.max().month == 12:
        last_year += 1

    results = []
    for table, gdp in zip(["NGDP", "NGDPD"], data):
        table_url = (
            f"https://www.imf.org/en/Publications/WEO/weo-database/"
            f"2024/October/weo-report?c=298,&s={table},&sy="
            f"{last_year - 1}&ey={last_year + 1}&ssm=0&scsm=1&scc=0&"
            f"ssd=1&ssc=0&sic=0&sort=country&ds=.&br=1"
        )
        imf_data = pd.to_numeric(pd.read_html(table_url)[0].iloc[0, [5, 6, 7]])
        imf_data = imf_data.reset_index(drop=True)
        fcast = (
            gdp.loc[[dt.datetime(last_year - 1, 12, 31)]]
            .multiply(imf_data.iloc[1])
            .divide(imf_data.iloc[0])
        )
        fcast = fcast.rename(
            index={dt.datetime(last_year - 1, 12, 31): dt.datetime(last_year, 12, 31)}
        )
        next_fcast = (
            gdp.loc[[dt.datetime(last_year - 1, 12, 31)]]
            .multiply(imf_data.iloc[2])
            .divide(imf_data.iloc[0])
        )
        next_fcast = next_fcast.rename(
            index={
                dt.datetime(last_year - 1, 12, 31): dt.datetime(last_year + 1, 12, 31)
            }
        )
        gdp = pd.concat([gdp, fcast, next_fcast], axis=0)
        results.append(gdp)

    output = pd.concat(results, axis=1)
    output = output.resample("QE-DEC").interpolate("linear").dropna(how="all")
    output = output.rename_axis(None)

    spanish_names = ["Producto bruto interno", "Producto bruto interno"]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "Millions",
        "seasonal_adjustment": None,
        "frequency": "QE-DEC",
        "time_series_type": "Flow",
        "cumulative_periods": 4,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    metadata.update_indicator_metadata_value(f"{name}_1", "currency", "USD")
    dataset = Dataset(name, output, metadata)

    return dataset


def industrial_production() -> Dataset:
    """Get industrial production data.

    Returns
    -------
    Monthly industrial production index : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=4, usecols="D:DR", na_values="(s)")
    weights = pd.read_csv(sources["weights"]).dropna(how="all")
    weights[["División", "Grupo", "Agrupación / Clase"]] = weights[
        ["División", "Grupo", "Agrupación / Clase"]
    ].astype(str)
    output = raw.dropna(how="all")
    output.index = pd.date_range(start="2018-01-31", freq="ME", periods=len(output))

    column_names = []
    for c in output.columns[2:]:
        c = str(c)
        match = weights.loc[weights["División"] == c, "Denominación"]
        prefix = "División: "
        if isinstance(match, pd.Series) and match.empty:
            prefix = "Grupo: "
            match = weights.loc[weights["Grupo"] == c, "Denominación"]
            if isinstance(match, pd.Series) and match.empty:
                prefix = "Clase: "
                match = weights.loc[weights["Agrupación / Clase"] == c, "Denominación"]
        try:
            match = match.iloc[0]
        except (AttributeError, IndexError):
            pass
        try:
            match = prefix + match.capitalize().strip()
        except AttributeError:
            match = c
        match = re.sub(r"[\(\)]", "-", match)
        column_names.append(match)

    output = output.apply(pd.to_numeric, errors="coerce").rename_axis(None)

    spanish_names = [
        "Industrias manufactureras",
        "Industrias manufactureras sin refinería",
    ] + column_names
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "2018=100",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def core_industrial_production(*args, **kwargs) -> Dataset:
    """
    Get total industrial production, industrial production excluding oil
    refinery and core industrial production.


    Returns
    -------
    Measures of industrial production : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    data_18 = load_dataset("industrial_production", *args, **kwargs).to_detailed()
    data_18 = data_18[
        [
            "Industrias manufactureras",
            "Industrias manufactureras sin refinería",
            "Clase: Elaboración de comidas y platos preparados; elaboración de otros productos alimenticios",
            "Clase: Fabricación de pasta de celulosa, papel y cartón",
        ]
    ].copy()
    data_18.columns = ["total", "ex-refinery", "other foods", "pulp"]

    data_18_weights = {"other foods": 0.0991, "pulp": 0.0907}
    data_18["core"] = data_18["ex-refinery"] - (
        data_18["other foods"] * data_18_weights["other foods"]
        + data_18["pulp"] * data_18_weights["pulp"]
    )
    data_18 = data_18[["total", "ex-refinery", "core"]]

    data_06 = pd.read_excel(
        sources["2006"], skiprows=6, usecols="B,D,F,CF,CX", na_values="(s)"
    ).dropna(how="all")
    data_06 = data_06.loc[~data_06.iloc[:, 0].str.contains("Prom")].iloc[:, 1:]
    data_06.columns = ["total", "ex-refinery", "other foods", "pulp"]
    data_06_weights = {
        "other foods": 0.3733 * 0.3107 * 0.7089,
        "pulp": 0.0184 * 0.4395,
    }  # https://www5.ine.gub.uy/documents/Estad%C3%ADsticasecon%C3%B3micas/SERIES%20Y%20OTROS/IVFIM/Ponderadores%20de%20VBP%202006.xls
    data_06["core"] = data_06["ex-refinery"] - (
        data_06["other foods"] * data_06_weights["other foods"]
        + data_06["pulp"] * data_06_weights["pulp"]
    )
    data_06 = data_06[["total", "ex-refinery", "core"]]
    data_06.index = pd.date_range(start="2002-01-31", freq="ME", periods=len(data_06))

    aux_index = list(dict.fromkeys(list(data_06.index) + list(data_18.index)))
    output = data_18.reindex(aux_index)
    for month in reversed(aux_index):
        if output.loc[month, :].isna().all():
            next_month = month + MonthEnd(1)
            output.loc[month, :] = (
                output.loc[next_month, :]
                * data_06.loc[month, :]
                / data_06.loc[next_month, :]
            )

    output = output.rename_axis(None)

    spanish_names = [
        "Industrias manufactureras",
        "Industrias manufactureras sin refinería",
        "Núcleo industrial",
    ]

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "2018=100",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata).rebase(
        start_date="2018-01-01", end_date="2018-12-31"
    )
    dataset.metadata.update_dataset_metadata({"unit": "2018=100"})
    dataset.transformed = False

    return dataset


def livestock_slaughter() -> Dataset:
    """Get weekly livestock slaughter data.

    Returns
    -------
    Weekly livestock slaughter : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    excel = pd.ExcelFile(sources["main"])
    cattle = pd.read_excel(excel, sheet_name="BOVINOS", skiprows=8, usecols="C:H")
    sheep = pd.read_excel(excel, sheet_name="OVINOS", skiprows=8, usecols="C:H")
    output = pd.concat([cattle, sheep], axis=1).fillna(0).astype(int)
    output.index = pd.date_range(start="2005-01-02", freq="W", periods=len(output))
    output = output.rename_axis(None)

    spanish_names = [
        "Novillos",
        "Vacas",
        "Vaquillonas",
        "Terneros",
        "Toros",
        "Total bovinos",
        "Borregos",
        "Capones",
        "Carneros",
        "Corderos",
        "Ovejas",
        "Total ovinos",
    ]

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": None,
        "inflation_adjustment": None,
        "unit": "Heads",
        "seasonal_adjustment": None,
        "frequency": "W",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def milk_shipments() -> Dataset:
    """Get monthly milk shipments from farms data.

    Returns
    -------
    Monhtly milk shipments from farms : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    r = httpx.get(sources["main"])
    url = re.findall(r'href="(.+\.xls)', r.text)[0]
    raw = pd.read_excel(
        url, sheet_name="Listado Datos", usecols="C:D", skiprows=4
    ).dropna()

    output = (
        raw.set_index(pd.date_range(start="2002-01-31", freq="ME", periods=len(raw)))
        / 1000
    )
    output = output.apply(pd.to_numeric)

    spanish_names = ["Remisión, litros", "Remisión, kilogramos"]

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": None,
        "inflation_adjustment": None,
        "unit": "Thousand liters",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    metadata.update_indicator_metadata_value(
        "milk_shipments_1", "unit", "Thousand kilograms"
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def diesel_sales() -> Dataset:
    """
    Get diesel sales by department data.

    This retrieval function requires the unrar binaries to be found in your
    system.

    Returns
    -------
    Monthly diesel dales : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    temp_rar = tempfile.NamedTemporaryFile(suffix=".rar").name
    with open(temp_rar, "wb") as f:
        r = httpx.get(sources["main"])
        rar_url = re.findall(
            r'(https?://[^"]*?gas%20oil[^"]*?\.rar)',
            r.text,
        )[0]
        f.write(httpx.get(rar_url).content)
    with tempfile.TemporaryDirectory() as temp_dir:
        patoolib.extract_archive(temp_rar, outdir=temp_dir, verbosity=-1)
        xls = [x for x in listdir(temp_dir) if x.endswith(".xls")][0]
        path_temp = path.join(temp_dir, xls)
        raw = pd.read_excel(
            path_temp, sheet_name="vta gas oil por depto", skiprows=2, usecols="C:W"
        )
        raw.index = pd.date_range(start="2004-01-31", freq="ME", periods=len(raw))
        raw.columns = list(raw.columns.str.replace("\n", " "))[:-1] + ["Total"]
        output = raw
    output = output.rename_axis(None)

    spanish_names = output.columns

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": None,
        "inflation_adjustment": None,
        "unit": "Cubic meters",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def gasoline_sales() -> Dataset:
    """
    Get gasoline sales by department data.

    This retrieval function requires the unrar binaries to be found in your
    system.

    Returns
    -------
    Monthly gasoline dales : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    temp_rar = tempfile.NamedTemporaryFile(suffix=".rar").name
    with open(temp_rar, "wb") as f:
        r = httpx.get(sources["main"])
        rar_url = re.findall(
            r'(https?://[^"]*?gasolinas[^"]*?\.rar)',
            r.text,
        )[0]
        f.write(httpx.get(rar_url).content)
    with tempfile.TemporaryDirectory() as temp_dir:
        patoolib.extract_archive(temp_rar, outdir=temp_dir, verbosity=-1)
        xls = [x for x in listdir(temp_dir) if x.endswith(".xls")][0]
        path_temp = path.join(temp_dir, xls)
        raw = pd.read_excel(
            path_temp, sheet_name="vta gasolinas por depto", skiprows=2, usecols="C:W"
        )
        raw.index = pd.date_range(start="2004-01-31", freq="ME", periods=len(raw))
        raw.columns = list(raw.columns.str.replace("\n", " "))[:-1] + ["Total"]
        output = raw
    output = output.rename_axis(None)

    spanish_names = output.columns

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": None,
        "inflation_adjustment": None,
        "unit": "Cubic meters",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def electricity_sales() -> Dataset:
    """
    Get electricity sales by sector data.

    This retrieval function requires the unrar binaries to be found in your
    system.

    Returns
    -------
    Monthly electricity dales : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    temp_rar = tempfile.NamedTemporaryFile(suffix=".rar").name
    with open(temp_rar, "wb") as f:
        r = httpx.get(sources["main"])
        rar_url = re.findall(
            r"https://www.gub.uy/ministerio-industria-energia-mineria/sites/ministerio-industria-energia-mineria/files/[0-9\-]+/Facturaci%C3%B3n%20de%20energ%C3%ADa%20el%C3%A9ctrica%20por%20sector.rar",
            r.text,
        )[0]
        f.write(httpx.get(rar_url).content)
    with tempfile.TemporaryDirectory() as temp_dir:
        patoolib.extract_archive(temp_rar, outdir=temp_dir, verbosity=-1)
        xls = [x for x in listdir(temp_dir) if x.endswith(".xls")][0]
        path_temp = path.join(temp_dir, xls)
        raw = pd.read_excel(path_temp, sheet_name="fact ee", skiprows=2, usecols="C:J")
        raw.index = pd.date_range(start="2000-01-31", freq="ME", periods=len(raw))
        raw.columns = raw.columns.str.capitalize()
        output = raw
    output = output.rename_axis(None)

    spanish_names = output.columns

    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids
    spanish_names = [{"es": x} for x in spanish_names]

    base_metadata = {
        "area": "Economic activity",
        "currency": None,
        "inflation_adjustment": None,
        "unit": "MWh",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Flow",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset



# import datetime as dt
# import tempfile
# import zipfile
# import os
# from io import BytesIO
# from os import path

# import pandas as pd
# import httpx
# from pandas.tseries.offsets import MonthEnd, QuarterEnd
# from dotenv import load_dotenv

# from econuy.transform import decompose, rebase
# from econuy.utils import metadata
# from econuy.utils.operations import get_download_sources, get_name_from_function


# load_dotenv()
# FRED_API_KEY = os.environ.get("FRED_API_KEY")


# def global_gdp() -> pd.DataFrame:
#     """Get seasonally adjusted real quarterly GDP for select countries.

#     Countries/aggregates are US, EU-27, Japan and China.

#     Returns
#     -------
#     Quarterly real GDP in seasonally adjusted terms : pd.DataFrame

#     """
#     name = get_name_from_function()
#     sources = get_download_sources(name)

#     if FRED_API_KEY is None:
#         raise ValueError(
#             "FRED_API_KEY not found. Get one at https://fredaccount.stlouisfed.org/apikeys and set it as an environment variable."
#         )

#     chn_y = dt.datetime.now().year + 1
#     chn_r = httpx.get(sources["chn_oecd"].format(f"{chn_y}-Q4"))
#     chn_datasets = (
#         pd.read_csv(BytesIO(chn_r.content))[
#             ["TIME_PERIOD", "TRANSFORMATION", "OBS_VALUE"]
#         ]
#         .pivot(index="TIME_PERIOD", columns="TRANSFORMATION", values="OBS_VALUE")
#         .rename_axis(None)
#         .rename_axis(None, axis=1)
#     )
#     chn_datasets.index = pd.to_datetime(
#         chn_datasets.index, format="mixed"
#     ) + QuarterEnd(0)
#     chn_qoq = chn_datasets.iloc[:, [0]]
#     chn_yoy = chn_datasets.iloc[:, [1]]
#     chn_obs = pd.read_csv(
#         sources["chn_obs"], index_col=0, parse_dates=[0], usecols=[0, 1]
#     )
#     chn_obs = chn_obs.loc[
#         (chn_obs.index > "2011-01-01") & (chn_obs.index < "2016-01-01")
#     ]
#     chn_yoy["volume"] = chn_obs
#     for row in reversed(range(len(chn_yoy.loc[chn_yoy.index < "2011-01-01"]))):
#         if pd.isna(chn_yoy.iloc[row, 1]):
#             chn_yoy.iloc[row, 1] = chn_yoy.iloc[row + 4, 1] / (
#                 1 + chn_yoy.iloc[row + 4, 0] / 100
#             )
#     chn_yoy = chn_yoy[["volume"]].loc[chn_yoy.index < "2016-01-01"]
#     metadata._set(chn_yoy)
#     chn_sa = decompose(
#         chn_yoy[["volume"]].loc[chn_yoy.index < "2016-01-01"],
#         component="seas",
#         method="x13",
#     )
#     chn_sa = pd.concat([chn_sa, chn_qoq], axis=1)
#     for row in range(len(chn_sa)):
#         if not pd.isna(chn_sa.iloc[row, 1]):
#             chn_sa.iloc[row, 0] = chn_sa.iloc[row - 1, 0] * (
#                 1 + chn_sa.iloc[row, 1] / 100
#             )
#     chn = chn_sa.iloc[:, [0]].div(10)

#     gdps = []
#     for series in ["GDPC1", "CLVMNACSCAB1GQEU272020", "JPNRGDPEXP"]:
#         r = httpx.get(
#             f"{sources['fred']}{series}&api_key=" f"{FRED_API_KEY}&file_type=json"
#         )
#         aux = pd.DataFrame.from_records(r.json()["observations"])
#         aux = aux[["date", "value"]].set_index("date")
#         aux.index = pd.to_datetime(aux.index)
#         aux.index = aux.index.shift(3, freq="ME") + MonthEnd(0)
#         aux.columns = [series]
#         aux = aux.apply(pd.to_numeric, errors="coerce")
#         if series == "GDPC1":
#             aux = aux.div(4)
#         elif series == "CLVMNACSCAB1GQEU272020":
#             aux = aux.div(1000)
#         gdps.append(aux)
#     gdps = pd.concat(gdps, axis=1)

#     output = pd.concat([gdps, chn], axis=1)
#     output.columns = ["Estados Unidos", "Unión Europea", "Japón", "China"]
#     output.rename_axis(None, inplace=True)

#     metadata._set(
#         output,
#         area="Global",
#         currency="USD",
#         inf_adj="Const.",
#         unit="Miles de millones",
#         seas_adj="SA",
#         ts_type="Flujo",
#         cumperiods=1,
#     )
#     metadata._modify_multiindex(
#         output, levels=[3], new_arrays=[["USD", "EUR", "JPY", "CNY"]]
#     )

#     return output


# def global_stock_markets() -> pd.DataFrame:
#     """Get stock market index data.

#     Indexes selected are S&P 500, Euronext 100, Nikkei 225 and Shanghai
#     Composite.

#     Returns
#     -------
#     Daily stock market index in USD : pd.DataFrame

#     """
#     name = get_name_from_function()
#     sources = get_download_sources(name)

#     yahoo = []
#     for series in ["spy", "n100", "nikkei", "sse"]:
#         aux = pd.read_csv(
#             sources[series].format(timestamp=dt.datetime.now().timestamp().__round__()),
#             index_col=0,
#             usecols=[0, 4],
#             parse_dates=True,
#         )
#         aux.columns = [series]
#         yahoo.append(aux)
#     output = pd.concat(yahoo, axis=1).interpolate(method="linear", limit_area="inside")
#     output.columns = [
#         "S&P 500",
#         "Euronext 100",
#         "Nikkei 225",
#         "Shanghai Stock Exchange Composite",
#     ]
#     output.rename_axis(None, inplace=True)
#     metadata._set(
#         output,
#         area="Global",
#         currency="USD",
#         inf_adj="No",
#         seas_adj="NSA",
#         ts_type="-",
#         cumperiods=1,
#     )
#     metadata._modify_multiindex(
#         output, levels=[3], new_arrays=[["USD", "EUR", "JPY", "CNY"]]
#     )
#     output = rebase(output, start_date="2019-01-02")

#     return output


# def global_policy_rates() -> pd.DataFrame:
#     """Get central bank policy interest rates data.

#     Countries/aggregates selected are US, Euro Area, Japan and China.

#     Returns
#     -------
#     Daily policy interest rates : pd.DataFrame

#     """
#     name = get_name_from_function()
#     sources = get_download_sources(name)

#     r = httpx.get(sources["main"])
#     temp_dir = tempfile.TemporaryDirectory()
#     with zipfile.ZipFile(BytesIO(r.content), "r") as f:
#         f.extractall(path=temp_dir.name)
#         path_temp = path.join(temp_dir.name, "WS_CBPOL_csv_row.csv")
#         raw = pd.read_csv(path_temp, index_col=0)
#     output = raw.loc[:, lambda x: x.columns.str.contains("D:Daily")]
#     output.columns = output.iloc[0]
#     output = (
#         output[["JP:Japan", "CN:China", "US:United States", "XM:Euro area"]]
#         .iloc[8:]
#         .dropna(how="all")
#     )
#     output.columns = ["Japón", "China", "Estados Unidos", "Eurozona"]
#     output = output.apply(pd.to_numeric, errors="coerce")

#     output = output[["Estados Unidos", "Eurozona", "Japón", "China"]]
#     output.index = pd.to_datetime(output.index)
#     output = output.interpolate(method="linear", limit_direction="forward")
#     output.rename_axis(None, inplace=True)

#     metadata._set(
#         output,
#         area="Global",
#         currency="USD",
#         inf_adj="No",
#         seas_adj="NSA",
#         unit="Tasa",
#         ts_type="-",
#         cumperiods=1,
#     )
#     metadata._modify_multiindex(
#         output, levels=[3], new_arrays=[["USD", "EUR", "JPY", "CNY"]]
#     )

#     return output


# # @retry(
# #     retry_on_exceptions=(HTTPError, URLError),
# #     max_calls_total=12,
# #     retry_window_after_first_call_in_seconds=60,
# # )
# # def long_rates() -> pd.DataFrame:
# #     """Get 10-year government bonds interest rates.

# #     Countries/aggregates selected are US, Germany, France, Italy, Spain
# #     United Kingdom, Japan and China.

# #     Returns
# #     -------
# #     Daily 10-year government bonds interest rates : pd.DataFrame

# #     """
# #     name = "global_long_rates"

# #     bonds = []
# #     r = httpx.get(f"{urls[name]['dl']['fred']}DGS10&api_key=" f"{FRED_API_KEY}&file_type=json")
# #     us = pd.DataFrame.from_records(r.json()["observations"])
# #     us = us[["date", "value"]].set_index("date")
# #     us.index = pd.to_datetime(us.index)
# #     us.columns = ["United States"]
# #     bonds.append(us.apply(pd.to_numeric, errors="coerce").dropna())

# #     for country, sid in zip(
# #         ["Germany", "France", "Italy", "Spain", "United Kingdom", "Japan", "China"],
# #         ["23693", "23778", "23738", "23806", "23673", "23901", "29227"],
# #     ):
# #         end_date_dt = dt.datetime(2000, 1, 1)
# #         start_date_dt = dt.datetime(2000, 1, 1)
# #         aux = []
# #         while end_date_dt < dt.datetime.now():
# #             end_date_dt = start_date_dt + dt.timedelta(days=5000)
# #             params = {
# #                 "curr_id": sid,
# #                 "smlID": str(randint(1000000, 99999999)),
# #                 "header": f"{country} 10-Year Bond Yield Historical Data",
# #                 "st_date": start_date_dt.strftime("%m/%d/%Y"),
# #                 "end_date": end_date_dt.strftime("%m/%d/%Y"),
# #                 "interval_sec": "Daily",
# #                 "sort_col": "date",
# #                 "sort_ord": "DESC",
# #                 "action": "historical_data",
# #             }
# #             r = httpx.post(
# #                 urls["global_long_rates"]["dl"]["main"], headers=investing_headers, data=params
# #             )
# #             aux.append(pd.read_html(r.content, match="Price", index_col=0,
# #                                     parse_dates=True, flavor="html.parser")[0])
# #             start_date_dt = end_date_dt + dt.timedelta(days=1)
# #         aux = pd.concat(aux, axis=0)[["Price"]].sort_index()
# #         aux.columns = [country]
# #         bonds.append(aux)

# #     output = bonds[0].join(bonds[1:], how="left")
# #     output = output.interpolate(method="linear", limit_area="inside")
# #     output.columns = [
# #         "Estados Unidos",
# #         "Alemania",
# #         "Francia",
# #         "Italia",
# #         "España",
# #         "Reino Unido",
# #         "Japón",
# #         "China",
# #     ]
# #     output.rename_axis(None, inplace=True)

# #     metadata._set(
# #         output,
# #         area="Global",
# #         currency="USD",
# #         inf_adj="No",
# #         seas_adj="NSA",
# #         unit="Tasa",
# #         ts_type="-",
# #         cumperiods=1,
# #     )
# #     metadata._modify_multiindex(
# #         output, levels=[3], new_arrays=[["USD", "EUR", "EUR", "EUR", "EUR", "GBP", "JPY", "CNY"]]
# #     )

# #     return output


# def global_nxr() -> pd.DataFrame:
#     """Get currencies data.

#     Selected currencies are the US dollar index, USDEUR, USDJPY and USDCNY.

#     Returns
#     -------
#     Daily currencies : pd.DataFrame

#     """
#     name = get_name_from_function()
#     sources = get_download_sources(name)

#     output = []
#     for series in ["dollar", "eur", "jpy", "cny"]:
#         aux = pd.read_csv(
#             sources[series].format(timestamp=dt.datetime.now().timestamp().__round__()),
#             index_col=0,
#             usecols=[0, 4],
#             parse_dates=True,
#         )
#         aux.columns = [series]
#         if series == "dollar":
#             aux.dropna(inplace=True)
#         output.append(aux)
#     output = (
#         output[0].join(output[1:]).interpolate(method="linear", limit_area="inside")
#     )
#     output.columns = ["Índice Dólar", "Euro", "Yen", "Renminbi"]
#     output.rename_axis(None, inplace=True)

#     metadata._set(
#         output,
#         area="Global",
#         currency="USD",
#         inf_adj="No",
#         seas_adj="NSA",
#         ts_type="-",
#         cumperiods=1,
#     )
#     metadata._modify_multiindex(
#         output,
#         levels=[3, 5],
#         new_arrays=[
#             ["USD", "EUR", "JPY", "CNY"],
#             ["Canasta/USD", "EUR/USD", "JPY/USD", "CNY/USD"],
#         ],
#     )

#     return output



import copy
import warnings
import json
from pathlib import Path
from datetime import datetime
from typing import List, Union, Optional, Literal, Dict

import pandas as pd
import numpy as np

from econuy.transform.change import _chg_diff
from econuy.transform.resample import _resample
from econuy.transform.rolling import _rolling
from econuy.transform.rebase import _rebase
from econuy.transform.convert import _convert_usd, _convert_gdp, _convert_real
from econuy.transform.decompose import _decompose


class DatasetConfig:
    def __init__(self, name: str) -> None:
        self.name = name
        self.load()

    def load(self) -> None:
        from econuy.utils.operations import REGISTRY

        dataset_config = REGISTRY[self.name]
        for key, value in dataset_config.items():
            setattr(self, key, value)

    def __repr__(self) -> str:
        return json.dumps(self.__dict__, indent=4)


class DatasetMetadata:
    def __init__(
        self,
        name: str,
        indicator_metadata: dict,
        created_at: Optional[datetime] = None,
        config: Optional[DatasetConfig] = None,
    ) -> None:
        self.name = name
        self.indicator_metadata = indicator_metadata
        self.created_at = created_at or datetime.now()
        self.config = config or DatasetConfig(name)

    def __getitem__(self, indicator) -> "DatasetMetadata":
        return self.__class__(
            name=self.name,
            indicator_metadata={indicator: self.indicator_metadata[indicator]},
        )

    def __setitem__(self, indicator, metadata) -> None:
        self.indicators[indicator] = metadata

    @property
    def indicator_ids(self) -> list:
        """
        Get the list of indicators in the metadata.

        Returns
        -------
        list
            The list of indicators.
        """
        return list(self.indicator_metadata.keys())

    @property
    def has_common_metadata(self) -> bool:
        """
        Check if all indicators have the same metadata.

        Returns
        -------
        bool
            True if all indicators have the same metadata, False otherwise.
        """
        metadata_wo_full_names = self._drop_full_names(self.indicator_metadata)
        indicator_metadatas = list(metadata_wo_full_names.values())
        if len(indicator_metadatas) < 2:
            return True
        else:
            reference = indicator_metadatas[0]
            return all(reference == metadata for metadata in indicator_metadatas[1:])

    @property
    def common_metadata_dict(self) -> dict:
        """
        Get the common metadata dictionary.

        Returns
        -------
        dict
            The metadata dictionary excluding keys that are not common to all indicators.
        """
        metadata_wo_full_names = self._drop_full_names(self.indicator_metadata)
        if len(metadata_wo_full_names) == 0:
            return {}

        first_metadata = metadata_wo_full_names[self.indicator_ids[0]]
        common_metadata = {}

        for key, value in first_metadata.items():
            if all(
                metadata_wo_full_names[indicator].get(key) == value
                for indicator in self.indicator_ids
            ):
                common_metadata[key] = value

        return common_metadata

    @staticmethod
    def _drop_full_names(indicator_metadata: dict) -> dict:
        """
        Drop full names from the metadata.

        Parameters
        ----------
        metadata : dict
            The metadata to process.

        Returns
        -------
        dict
            The metadata with full names dropped.
        """
        return {
            indicator: {
                meta_name: meta
                for meta_name, meta in single_indicator_metadata.items()
                if "name" not in meta_name
            }
            for indicator, single_indicator_metadata in indicator_metadata.items()
        }

    def update_indicator_metadata(
        self, indicator: str, single_indicator_metadata: dict
    ) -> "DatasetMetadata":
        """
        Update the metadata for a specific indicator.

        Parameters
        ----------
        indicator : str
            The indicator to update.
        single_indicator_metadata : dict
            The metadata to update with.

        Returns
        -------
        Metadata
            The updated metadata.
        """
        self.indicator_metadata[indicator].update(single_indicator_metadata)
        return self

    def update_indicator_metadata_value(
        self,
        indicator: str,
        key: str,
        value: str,
    ) -> "DatasetMetadata":
        """
        Update the metadata for a specific indicator.

        Parameters
        ----------
        indicator : str
            The indicator to update.
        key : str
            The key to update.
        value : str
            The value to update with.

        Returns
        -------
        Metadata
            The updated metadata.
        """
        self.indicator_metadata[indicator][key] = value
        return self

    def update_dataset_metadata(self, indicator_metadata: dict) -> "DatasetMetadata":
        """
        Update the metadata for all indicators.

        Parameters
        ----------
        indicator_metadata : dict
            The new metadata to update with.

        Returns
        -------
        Metadata
            The updated metadata.
        """
        for indicator in self.indicator_metadata:
            self.indicator_metadata[indicator].update(indicator_metadata)
        return self

    def add_transformation_step(self, transformation: dict) -> "DatasetMetadata":
        """
        Add a transformation step to the metadata.

        Parameters
        ----------
        transformation : dict
            The transformation step to add.

        Returns
        -------
        Metadata
            The updated metadata.
        """
        for indicator in self.indicator_metadata:
            self.indicator_metadata[indicator]["transformations"].append(transformation)
        return self

    def copy(self) -> "DatasetMetadata":
        """
        Create a copy of the metadata.

        Returns
        -------
        Metadata
            The copied metadata.
        """
        return copy.deepcopy(self)

    def to_dict(self) -> Dict:
        d = self.__dict__.copy()
        d["created_at"] = d["created_at"].isoformat()
        d["config"] = self.config.__dict__
        return d

    def save(self, name: str, data_dir: Union[str, Path, None] = None) -> None:
        from econuy.utils.operations import get_data_dir

        data_dir = data_dir or get_data_dir()
        data_dir = Path(data_dir)
        data_dir.mkdir(parents=True, exist_ok=True)
        for_json = self.to_dict()
        with open(data_dir / f"{name}_metadata.json", "w") as f:
            json.dump(for_json, f, indent=4)
        return

    @staticmethod
    def cast_metadata(
        indicator_metadata: dict, indicator_ids: list, full_names: list
    ) -> dict:
        # TODO: Improve this hack
        return {
            name: {"names": full_name} | indicator_metadata
            for name, full_name in zip(indicator_ids, full_names)
        }

    @classmethod
    def from_cast(
        cls, name: str, base_metadata: dict, indicator_ids: list, indicator_names: list
    ) -> "DatasetMetadata":
        """
        Create a metadata instance from a casted metadata.

        Parameters
        ----------
        base_metadata : dict
            The base metadata.
        names : list
            The names of the indicators.
        full_names : list
            The full names of the indicators.

        Returns
        -------
        Metadata
            The created metadata instance.
        """
        indicator_metadata = cls.cast_metadata(
            base_metadata, indicator_ids, indicator_names
        )
        return cls(name, indicator_metadata)

    @classmethod
    def from_metadatas(
        cls, name: str, metadatas: List["DatasetMetadata"]
    ) -> "DatasetMetadata":
        """
        Create a metadata instance from a list of metadatas.

        Parameters
        ----------
        metadatas : list
            The list of metadatas.

        Returns
        -------
        Metadata
            The created metadata instance.
        """
        metadatas_dict = {
            k: v for d in metadatas for k, v in d.indicator_metadata.items()
        }
        return cls(name, metadatas_dict)

    @classmethod
    def from_json(cls, path: Union[str, Path]) -> "DatasetMetadata":
        """
        Create a metadata instance from a JSON file.

        Parameters
        ----------
        path : str or Path
            The path to the JSON file.

        Returns
        -------
        Metadata
            The created metadata instance.
        """
        with open(path, "r") as f:
            metadata_dict = json.load(f)
        metadata_dict["created_at"] = datetime.fromisoformat(
            metadata_dict["created_at"]
        )
        metadata_dict["config"] = DatasetConfig(metadata_dict["name"])
        return cls(**metadata_dict)

    def __repr__(self) -> str:
        return "\n".join(
            [
                f"Name: {self.name}",
                f"Created at: {self.created_at}",
                f"Indicator metadata: {self.indicator_metadata}",
            ]
        )


class Dataset:
    """
    A class to represent a collection of economic data.

    Parameters
    ----------
    data : pd.DataFrame
        The economic data.
    metadata : Metadata
        The metadata of the data.
    name : str
        The name of the dataset.

    Returns
    -------
    None

    See Also
    --------
    :class:`pd.DataFrame`
    :class:`Metadata`
    """

    def __init__(
        self,
        name: str,
        data: pd.DataFrame,
        metadata: DatasetMetadata,
        transformed: bool = False,
    ) -> None:
        """
        Initialize the dataset.

        Parameters
        ----------
        name : str
            The name of the dataset.
        data : pd.DataFrame
            The economic data.
        metadata : Metadata
            The metadata of the data.
        transformed : bool
            Whether the data has been transformed.

        Returns
        -------
        None
        """
        self.data = data
        self.metadata = metadata
        self.name = name
        self.transformed = transformed
        self.indicators = self.metadata.indicator_ids

    def validate(self) -> None:
        """
        Validate the dataset.

        Raises
        ------
        AssertionError
            If the number of indicators does not match the number of columns in the data.
            If any of the indicators are not in the data.
            If the index of the data is not a DatetimeIndex.
            If the data contains non-numeric values.

        """
        assert len(self.indicators) == len(self.data.columns)
        assert all(indicator in self.data.columns for indicator in self.indicators)
        assert isinstance(self.data.index, pd.DatetimeIndex)
        assert self.data.dtypes.apply(pd.api.types.is_numeric_dtype).all()

    def to_detailed(self, language: str = "es") -> pd.DataFrame:
        """
        Rename the data using the metadata.

        Parameters
        ----------
        language : str, default "es"
            The language to use for the metadata.

        Returns
        -------
        pd.DataFrame
            The data with the indicators renamed.

        """
        column_metadatas = {
            indicator: {
                "name": self.metadata.indicator_metadata[indicator]["names"][language]
            }
            for indicator in self.indicators
        }
        for ind, col_meta in column_metadatas.items():
            col_meta.update(self.metadata.indicator_metadata[ind])
            col_meta.update({"id": ind})
            col_meta.pop("names")
            col_meta.pop(
                "transformations", None
            )  # TODO: Add transformations to metadata
        col_names = [
            x.replace("_", " ").capitalize()
            for x in column_metadatas[self.indicators[0]].keys()
        ]
        columns = pd.MultiIndex.from_tuples(
            [tuple(column_metadatas[k].values()) for k in column_metadatas.keys()],
            names=col_names,
        )
        detailed_data = self.data.copy()
        detailed_data.columns = columns
        return detailed_data

    def to_named(self, language: str = "es") -> pd.DataFrame:
        """
        Rename the data using the metadata.

        Parameters
        ----------
        language : str, default "es"
            The language to use for the metadata.

        Returns
        -------
        pd.DataFrame
            The data with the indicators renamed.

        """
        column_metadatas = {
            indicator: self.metadata.indicator_metadata[indicator]["names"][language]
            for indicator in self.indicators
        }
        named_data = self.data.copy()
        named_data.columns = [column_metadatas[ind] for ind in self.indicators]
        return named_data

    def to_json(self) -> dict:
        """
        Convert the dataset to a valid JSON dictionary.

        Returns
        -------
        dict
            A JSON representation of the dataset.

        """
        data = self.data.copy()
        data.index = data.index.astype(str)
        data = data.replace([np.inf, -np.inf], np.nan)
        data = data.astype(object).where(pd.notnull(data), None)

        metadata = self.metadata.to_dict()
        metadata.pop("config")
        return {
            "name": self.name,
            "data": data.to_dict(),
            "metadata": metadata,
            "transformed": self.transformed,
        }

    def save(
        self, data_dir: Union[str, Path, None] = None, name: Optional[str] = None
    ) -> None:
        """
        Save the dataset to a directory.

        Parameters
        ----------
        data_dir : str or Path
            The directory to save the dataset to.
        name : str, default None
            The name to save the dataset as without suffixes.

        Returns
        -------
        None

        """
        from econuy.utils.operations import get_data_dir

        data_dir = data_dir or get_data_dir()
        data_dir = Path(data_dir)
        data_dir.mkdir(parents=True, exist_ok=True)
        name = name or (f"{self.name}_transformed" if self.transformed else self.name)
        self.data.to_csv(data_dir / f"{name}.csv")
        self.metadata.save(name, data_dir)
        return

    def infer_frequency(self) -> Optional[pd.Timedelta]:
        """
        Infer the frequency of the data.

        Returns
        -------
        Optional[pd.Timedelta]
            The inferred frequency of the data.

        """
        try:
            inferred_freq = pd.infer_freq(self.data.index)
        except ValueError:
            warnings.warn(
                "ValueError: Need at least 3 dates to infer frequency. "
                "Setting to 'None'.",
                UserWarning,
                stacklevel=2,
            )
            inferred_freq = None
        if inferred_freq is None:
            warnings.warn(
                "Metadata: frequency could not be inferred "
                "from the index. Setting to 'None'.",
                UserWarning,
                stacklevel=2,
            )
            inferred_freq = None
        return inferred_freq

    def call_pandas_method(self, method: str, *args, **kwargs) -> "Dataset":
        output = self.__class__(
            data=self.data.__getattribute__(method)(*args, **kwargs),
            metadata=self.metadata,
            name=self.name,
            transformed=self.transformed,
        )
        return output

    def __getitem__(self, indicators: Union[str, List[str]]) -> "Dataset":
        indicators = [indicators] if isinstance(indicators, str) else indicators
        metadata_dict = {i: self.metadata.indicator_metadata[i] for i in indicators}
        return self.__class__(
            data=self.data[indicators],
            metadata=DatasetMetadata(self.name, metadata_dict),
            name=self.name,
            transformed=self.transformed,
        )

    def select(
        self,
        ids: Union[str, List[str], None] = None,
        names: Union[str, List[str], None] = None,
        language: str = "es",
    ) -> "Dataset":
        assert (
            ids is not None or names is not None
        ), "Either 'ids' or 'names' must be provided."
        assert (
            ids is None or names is None
        ), "Only one of 'ids' or 'names' can be provided."

        if ids is None:
            names = [names] if isinstance(names, str) else names
            ids = [
                k
                for k, v in self.metadata.indicator_metadata.items()
                if v["names"][language] in names
            ]
        return self.__getitem__(ids)

    def filter(
        self,
        start_date: Union[str, datetime, None] = None,
        end_date: Union[str, datetime, None] = None,
    ) -> "Dataset":
        return self.__class__(
            data=self.data.loc[start_date:end_date],
            metadata=self.metadata,
            name=self.name,
            transformed=self.transformed,
        )

    def __repr__(self) -> str:
        return "\n".join(
            [
                f"Name: {self.name}",
                f"Indicators: {self.indicators}",
            ]
        )

    def resample(
        self,
        rule: Union[pd.DateOffset, pd.Timedelta, str],
        operation: Literal["sum", "mean", "last", "upsample"] = "sum",
        interpolation: str = "linear",
    ) -> "Dataset":
        """
        Wrapper for the `resample method <https://pandas.pydata.org/pandas-docs
        stable/reference/api/pandas.DataFrame.resample.html>`_ in Pandas that
        integrates with econuy dataframes' metadata.

        Trim partial bins, i.e. do not calculate the resampled
        period if it is not complete, unless the input dataframe has no defined
        frequency, in which case no trimming is done.

        Parameters
        ----------
        rule : pd.DateOffset, pd.Timedelta or str
            Target frequency to resample to. See
            `Pandas offset aliases <https://pandas.pydata.org/pandas-docs/stable/
            user_guide/timeseries.html#offset-aliases>`_
        operation : {'sum', 'mean', 'last', 'upsample'}
            Operation to use for resampling.
        interpolation : str, default 'linear'
            Method to use when missing data are produced as a result of
            resampling, for example when upsampling to a higher frequency. See
            `Pandas interpolation methods <https://pandas.pydata.org/pandas-docs
            /stable/reference/api/pandas.Series.interpolate.html>`_

        Returns
        -------
        ``Dataset``

        Raises
        ------
        ValueError
            If ``operation`` is not one of available options.
        ValueError
            If the input dataframe's columns do not have the appropiate levels.

        Warns
        -----
        UserWarning
            If input frequencies cannot be assigned a numeric value, preventing
            incomplete bin trimming.

        """
        if operation not in ["sum", "mean", "upsample", "last"]:
            raise ValueError("Invalid 'operation' option.")

        if self.metadata.has_common_metadata:
            transformed, new_metadata = _resample(
                data=self.data,
                metadata=self.metadata,
                rule=rule,
                operation=operation,
                interpolation=interpolation,
            )
        else:
            transformed = []
            new_metadatas = []
            for column_name in self.data.columns:
                n_dataset = self[column_name]
                transformed_col, new_metadata = _resample(
                    data=n_dataset.data,
                    metadata=n_dataset.metadata,
                    rule=rule,
                    operation=operation,
                    interpolation=interpolation,
                )
                transformed.append(transformed_col)
                new_metadatas.append(new_metadata)
            transformed = pd.concat(transformed, axis=1)
            new_metadata = DatasetMetadata.from_metadatas(self.name, new_metadatas)

        inferred_frequency = pd.infer_freq(transformed.index)
        new_metadata.update_dataset_metadata({"frequency": inferred_frequency})
        output = self.__class__(
            data=transformed, metadata=new_metadata, name=self.name, transformed=True
        )
        return output

    def rolling(
        self, window: int, operation: Literal["sum", "mean"] = "sum"
    ) -> "Dataset":
        """
        Wrapper for the `rolling method <https://pandas.pydata.org/pandas-docs/
        stable/reference/api/pandas.DataFrame.rolling.html>`_ in Pandas that
        integrates with econuy dataframes' metadata.

        If ``periods`` is ``None``, try to infer the frequency and set ``periods``
        according to the following logic: ``{'YE-DEC': 1, 'QE-DEC': 4, 'ME': 12}``, that
        is, each period will be calculated as the sum or mean of the last year.

        Parameters
        ----------
        window : int, default None
            How many periods the window should cover.
        operation : {'sum', 'mean'}
            Operation used to calculate rolling windows.

        Returns
        -------
        ``Dataset``

        Raises
        ------
        ValueError
            If ``operation`` is not one of available options.
        ValueError
            If the input dataframe's columns do not have the appropiate levels.

        Warns
        -----
        UserWarning
            If the input dataframe is a stock time series, for which rolling
            operations are not recommended.

        """
        if operation not in ["sum", "mean"]:
            raise ValueError("Invalid 'operation' option.")

        if self.metadata.has_common_metadata:
            transformed, new_metadata = _rolling(
                data=self.data,
                metadata=self.metadata,
                window=window,
                operation=operation,
            )

        else:
            transformed = []
            new_metadatas = []
            for column_name in self.data.columns:
                n_dataset = self[column_name]
                transformed_col, new_metadata = _rolling(
                    data=n_dataset.data,
                    metadata=n_dataset.metadata,
                    window=window,
                    operation=operation,
                )
                transformed.append(transformed_col)
                new_metadatas.append(new_metadata)
            transformed = pd.concat(transformed, axis=1)
            new_metadata = DatasetMetadata.from_metadatas(self.name, new_metadatas)
        output = self.__class__(
            data=transformed,
            metadata=new_metadata,
            name=self.name,
            transformed=True,
        )
        return output

    def chg_diff(
        self,
        operation: Literal["chg", "diff"] = "chg",
        period: Literal["last", "inter", "annual"] = "last",
    ) -> "Dataset":
        """Wrapper for the `pct_change <https://pandas.pydata.org/pandas-docs/stable/
        reference/api/pandas.DataFrame.pct_change.html>`_ and `diff <https://pandas
        .pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.diff.html>`_
        Pandas methods.

        Calculate percentage change or difference for dataframes. The ``period``
        argument takes into account the frequency of the dataframe, i.e.,
        ``inter`` (for interannual) will calculate pct change/differences with
        ``periods=4`` for quarterly frequency, but ``periods=12`` for monthly
        frequency.

        Parameters
        ----------
        operation : {'chg', 'diff'}
            ``chg`` for percent change or ``diff`` for differences.
        period : {'last', 'inter', 'annual'}
            Period with which to calculate change or difference. ``last`` for
            previous period (last month for monthly data), ``inter`` for same
            period last year, ``annual`` for same period last year but taking
            annual sums.

        Returns
        -------
        ``Dataset``

        Raises
        ------
        ValueError
            If the dataframe is not of frequency ``ME`` (month), ``QE`` or
            ``QE-DEC`` (quarter), or ``YE`` or ``YE-DEC`` (year).
        ValueError
            If the ``operation`` parameter does not have a valid argument.
        ValueError
            If the ``period`` parameter does not have a valid argument.
        ValueError
            If the input dataframe's columns do not have the appropiate levels.

        """
        if operation not in ["chg", "diff"]:
            raise ValueError("Invalid 'operation' option.")
        if period not in ["last", "inter", "annual"]:
            raise ValueError("Invalid 'period' option.")

        if self.metadata.has_common_metadata:
            transformed, new_metadata = _chg_diff(
                data=self.data,
                metadata=self.metadata,
                operation=operation,
                period=period,
            )

        else:
            transformed = []
            new_metadatas = []
            for column_name in self.data.columns:
                n_dataset = self[column_name]
                transformed_col, new_metadata = _chg_diff(
                    data=n_dataset.data,
                    metadata=n_dataset.metadata,
                    operation=operation,
                    period=period,
                )
                transformed.append(transformed_col)
                new_metadatas.append(new_metadata)
            transformed = pd.concat(transformed, axis=1)
            new_metadata = DatasetMetadata.from_metadatas(self.name, new_metadatas)
        output = self.__class__(
            data=transformed,
            metadata=new_metadata,
            name=self.name,
            transformed=True,
        )
        return output

    def rebase(
        self,
        start_date: Union[str, datetime],
        end_date: Union[str, datetime, None] = None,
        base: float = 100.0,
    ) -> "Dataset":
        """Rebase dataset to a date or range of dates.

        Parameters
        ----------
        start_date : string or datetime.datetime
            Date to which series will be rebased.
        end_date : string or datetime.datetime, default None
            If specified, series will be rebased to the average between
            ``start_date`` and ``end_date``.
        base : float, default 100
            Float for which ``start_date`` == ``base`` or average between
            ``start_date`` and ``end_date`` == ``base``.

        Returns
        -------
        ``Dataset``

        """

        if self.metadata.has_common_metadata:
            transformed, new_metadata = _rebase(
                data=self.data,
                metadata=self.metadata,
                start_date=start_date,
                end_date=end_date,
                base=base,
            )

        else:
            transformed = []
            new_metadatas = []
            for column_name in self.data.columns:
                n_dataset = self[column_name]
                transformed_col, new_metadata = _rebase(
                    data=n_dataset.data,
                    metadata=n_dataset.metadata,
                    start_date=start_date,
                    end_date=end_date,
                    base=base,
                )
                transformed.append(transformed_col)
                new_metadatas.append(new_metadata)
            transformed = pd.concat(transformed, axis=1)
            new_metadata = DatasetMetadata.from_metadatas(self.name, new_metadatas)
        output = self.__class__(
            data=transformed,
            metadata=new_metadata,
            name=self.name,
            transformed=True,
        )
        return output

    def convert(
        self,
        flavor: Literal["usd", "real", "gdp"],
        start_date: Union[str, datetime, None] = None,
        end_date: Union[str, datetime, None] = None,
        error_handling: Literal["raise", "coerce", "ignore"] = "raise",
    ) -> "Dataset":
        """Convert dataset from UYU to USD, from UYU to real UYU or
        from UYU/USD to % GDP.

        ``flavor=usd``: Convert a dataset from Uruguayan pesos to US dollars. Takes into
        account whether the input datasets is flow or stock, in order to
        choose end of period or monthly average NXR. Also take into account the
        input dataframe's frequency and whether columns represent rolling averages
        or sums.

        ``flavor=real``: Convert a dataset columns to real prices. Takes into account the
        input datasets's frequency and whether
        columns represent rolling averages or sums. Allow choosing a single period,
        a range of dates or no period as a base (i.e., period for which the
        average/sum of input dataframe and output dataframe is the same).

        ``flavor=gdp``: Convert a dataset to percentage of GDP. Takes into account the
        input dataset's currency for chossing UYU or USD GDP. If frequency of input dataset is
        higher than quarterly, GDP will be upsampled and linear interpolation will
        be performed to complete missing data.
        If input dataframe's "cumulative_periods" level is not 12 for monthly frequency or 4
        for quarterly frequency, calculate rolling input dataframe.

        In all cases, if input dataframe's frequency is higher than monthly
        (daily, business, etc.), resample to monthly frequency.

        Parameters
        ----------
        flavor : str
            ``usd`` for USD, ``real`` for real UYU, ``gdp`` for % GDP.
        start_date : str, datetime.date or None, default None
            Only used if ``flavor=real``. If set to a date-like string or a
            date, and ``end_date`` is None, the base period will be
            ``start_date``.
        end_date : str, datetime.date or None, default None
            Only used if ``flavor=real``. If ``start_date`` is set, calculate
            so that the data is in constant prices of ``start_date-end_date``.
        error_handling : {"raise", "coerce", "ignore"}, default "raise"
            What to do when the input dataset can't be converted. Coercion will set to np.nan,
            while "ignore" is a no-op. If "raise", will raise an error.

        Returns
        -------
        ``Dataset``
        """
        assert flavor in ["usd", "real", "gdp"], "Invalid 'flavor' option."
        funcs = {"usd": _convert_usd, "real": _convert_real, "gdp": _convert_gdp}
        func = funcs[flavor]
        kwargs = (
            {"start_date": start_date, "end_date": end_date} if flavor == "real" else {}
        )

        if self.metadata.has_common_metadata:
            transformed, new_metadata = func(
                data=self.data,
                metadata=self.metadata,
                error_handling=error_handling,
                **kwargs,
            )

        else:
            transformed = []
            new_metadatas = []
            for column_name in self.data.columns:
                n_dataset = self[column_name]
                transformed_col, new_metadata = func(
                    data=n_dataset.data,
                    metadata=n_dataset.metadata,
                    error_handling=error_handling,
                    **kwargs,
                )
                transformed.append(transformed_col)
                new_metadatas.append(new_metadata)
            transformed = pd.concat(transformed, axis=1)
            new_metadata = DatasetMetadata.from_metadatas(self.name, new_metadatas)
        output = self.__class__(
            data=transformed,
            metadata=new_metadata,
            name=self.name,
            transformed=True,
        )
        return output

    def decompose(
        self,
        method: Literal["x13", "loess", "mloess", "moving_averages"] = "x13",
        fallback: Literal["loess", "mloess", "moving_averages"] = "loess",
        component: Literal["t-c", "sa"] = "sa",
        fn_kwargs: Optional[dict] = None,
        ignore_warnings: bool = True,
        error_handling: Literal["raise", "coerce", "ignore"] = "raise",
    ) -> "Dataset":
        """Rebase dataset to a date or range of dates.

        Parameters
        ----------
        method : {'x13', 'loess', 'mloess', 'moving_averages'}, default 'x13'
            Method to use for decomposition.
        fallback : {'loess', 'mloess', 'moving_averages'}, default 'loess'
            Fallback method to use if the main method fails.
        component : {'t-c', 'sa'}, default 'sa'
            Component to return. 't-c' for trend-cycle, 'sa' for seasonally adjusted.
        fn_kwargs : dict, default None
            Additional keyword arguments to pass to the decomposition function.
        ignore_warnings : bool, default True
            Whether to ignore warnings.
        error_handling : {"raise", "coerce", "ignore"}, default "raise"
            What to do when the input dataset can't be converted. Coercion will set to np.nan,

        Returns
        -------
        ``Dataset``

        """
        assert method in [
            "x13",
            "loess",
            "mloess",
            "moving_averages",
        ], "Invalid 'method' option."
        assert component in ["t-c", "sa"], "Invalid 'component' option."

        fn_kwargs = fn_kwargs or {}

        if self.metadata.has_common_metadata:
            transformed, new_metadata = _decompose(
                data=self.data,
                metadata=self.metadata,
                method=method,
                fallback=fallback,
                component=component,
                fn_kwargs=fn_kwargs,
                ignore_warnings=ignore_warnings,
                error_handling=error_handling,
            )

        else:
            transformed = []
            new_metadatas = []
            for column_name in self.data.columns:
                n_dataset = self[column_name]
                transformed_col, new_metadata = _decompose(
                    data=n_dataset.data,
                    metadata=n_dataset.metadata,
                    method=method,
                    fallback=fallback,
                    component=component,
                    fn_kwargs=fn_kwargs,
                    ignore_warnings=ignore_warnings,
                    error_handling=error_handling,
                )
                transformed.append(transformed_col)
                new_metadatas.append(new_metadata)
            transformed = pd.concat(transformed, axis=1)
            new_metadata = DatasetMetadata.from_metadatas(self.name, new_metadatas)
        output = self.__class__(
            data=transformed,
            metadata=new_metadata,
            name=self.name,
            transformed=True,
        )
        return output



import pandas as pd
from pandas.tseries.offsets import MonthEnd

from econuy.base import Dataset, DatasetMetadata
from econuy import load_dataset
from econuy.utils.operations import (
    get_name_from_function,
    get_download_sources,
    get_names_and_ids,
    get_base_metadata,
)


def labor_rates_gender() -> Dataset:
    """Get labor market data (LFPR, employment and unemployment).

    Returns
    -------
    Monthly participation, employment and unemployment rates : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=7).dropna(axis=0, thresh=2)
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2006-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")
    output.columns = [
        "Tasa de actividad: total",
        "Tasa de actividad: hombres",
        "Tasa de actividad: mujeres",
        "Tasa de empleo: total",
        "Tasa de empleo: hombres",
        "Tasa de empleo: mujeres",
        "Tasa de desempleo: total",
        "Tasa de desempleo: hombres",
        "Tasa de desempleo: mujeres",
    ]

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Labor market",
        "currency": None,
        "inflation_adjustment": None,
        "unit": "Rate",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def activity_region() -> Dataset:
    """Get activity rates by region (LFPR).

    Returns
    -------
    Monthly participation rates : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=8).dropna(axis=0, thresh=2)
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2006-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids, spanish_names = get_names_and_ids(name, "es")
    output.columns = ids

    base_metadata = get_base_metadata(name)
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def employment_region() -> Dataset:
    """Get employment rates by region.

    Returns
    -------
    Monthly employment rates : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=8).dropna(axis=0, thresh=2)
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2006-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids, spanish_names = get_names_and_ids(name, "es")
    output.columns = ids

    base_metadata = get_base_metadata(name)
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def unemployment_region() -> Dataset:
    """Get unemployment rates by region.

    Returns
    -------
    Monthly unemployment rates : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=8).dropna(axis=0, thresh=2)
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2006-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids, spanish_names = get_names_and_ids(name, "es")
    output.columns = ids

    base_metadata = get_base_metadata(name)
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def employment_age() -> Dataset:
    """Get employment rates by age group.

    Returns
    -------
    Monthly employment rates : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=7).dropna(axis=0, thresh=2)
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2006-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids, spanish_names = get_names_and_ids(name, "es")
    output.columns = ids

    base_metadata = get_base_metadata(name)
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def unemployment_contributions() -> Dataset:
    """Get contributions to unemployment rates.
    Returns
    -------
    Monthly unemployment rates : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=9).dropna(axis=0, thresh=2)
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2006-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids, spanish_names = get_names_and_ids(name, "es")
    output.columns = ids

    base_metadata = get_base_metadata(name)
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def unemployment_characteristics() -> Dataset:
    """Get unemployment rates by characteristics.

    Returns
    -------
    Monthly unemployment rates : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=8).dropna(axis=0, thresh=2)
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2006-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids, spanish_names = get_names_and_ids(name, "es")
    output.columns = ids

    base_metadata = get_base_metadata(name)
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def unemployment_conditions() -> Dataset:
    """Get unemployment rates by conditions.

    Returns
    -------
    Monthly unemployment rates : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(
        sources["main"], skiprows=9, na_values=[".."], usecols="A:I"
    ).dropna(axis=0, thresh=2)
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2006-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids, spanish_names = get_names_and_ids(name, "es")
    output.columns = ids

    base_metadata = get_base_metadata(name)
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def unemployment_duration() -> Dataset:
    """Get average duration of unemployment in weeks.

    Returns
    -------
    Monthly average duration of unemployment : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=9, usecols="A,J").dropna(
        axis=0, thresh=2
    )
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2006-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids, spanish_names = get_names_and_ids(name, "es")
    output.columns = ids

    base_metadata = get_base_metadata(name)
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def employment_characteristics() -> Dataset:
    """Get employment rates by characteristics.

    Returns
    -------
    Monthly employment rates : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=8).dropna(axis=0, thresh=2)
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2006-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids, spanish_names = get_names_and_ids(name, "es")
    output.columns = ids

    base_metadata = get_base_metadata(name)
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def nominal_wages() -> Dataset:
    """Get nominal general, public and private sector wages data

    Returns
    -------
    Monthly wages separated by public and private sector : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    historical = pd.read_excel(sources["historical"], skiprows=8, usecols="A:B")
    current = pd.read_excel(sources["current"], skiprows=8, usecols="A,C:D")
    historical = historical.dropna(how="any").set_index("Unnamed: 0")
    current = current.dropna(how="any").set_index("Unnamed: 0")
    output = pd.concat([historical, current], axis=1)
    output.index = output.index + MonthEnd(1)
    output.columns = [
        "Índice medio de salarios",
        "Índice medio de salarios privados",
        "Índice medio de salarios públicos",
    ]
    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Labor market",
        "currency": "UYU",
        "inflation_adjustment": None,
        "unit": "2008-07=100",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def employment_sector() -> Dataset:
    """Get employment data by sector.

    Returns
    -------
    Monthly employment data by sector : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"], skiprows=7).dropna(axis=0, thresh=2)
    output = raw[~raw["Unnamed: 0"].str.contains("-|/|Total", regex=True)]
    output.index = pd.date_range(start="2011-01-31", periods=len(output), freq="ME")
    output = output.drop(columns="Unnamed: 0")

    output = output.apply(pd.to_numeric, errors="coerce")
    output = output.rename_axis(None)

    ids, spanish_names = get_names_and_ids(name, "es")
    output.columns = ids

    base_metadata = get_base_metadata(name)
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def hours_worked() -> Dataset:
    """Get average hours worked data.

    Returns
    -------
    Monthly hours worked : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    raw = pd.read_excel(sources["main"]).dropna(axis=0, thresh=2)

    output = raw[~raw.iloc[:, 0].str.contains("-|/|Total|Año", regex=True)].iloc[:, 1:]
    output.index = pd.date_range(start="2011-01-31", periods=len(output), freq="ME")
    output.columns = [
        "Total",
        "Industrias manufactureras",
        "Electricidad, gas, agua y saneamiento",
        "Construcción",
        "Comercio",
        "Transporte y almacenamiento",
        "Alojamiento y servicios de comidas",
        "Información y comunicación",
        "Actividades financieras",
        "Actividades inmobiliarias y administrativas",
        "Actividades profesionales",
        "Administración pública y seguridad social",
        "Enseñanza",
        "Salud",
        "Arte y otros servicios",
        "Act. de hogares como empleadores",
        "Agro, forestación, pesca y minería",
    ]

    output = output.rename_axis(None)
    output = output.apply(pd.to_numeric, errors="coerce")

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Labor market",
        "currency": None,
        "inflation_adjustment": None,
        "unit": "Hours per week",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset


def labor_rates_persons(*args, **kwargs) -> Dataset:
    """
    Get labor data, both rates and persons. Extends national data between 1991
    and 2005 with data for jurisdictions with more than 5,000 inhabitants.

    Returns
    -------
    Labor market data : Dataset

    """
    name = get_name_from_function()
    sources = get_download_sources(name)

    rates = load_dataset("labor_rates", *args, **kwargs).to_named()
    rates = rates.loc[
        :,
        [
            "Tasa de actividad: total",
            "Tasa de empleo: total",
            "Tasa de desempleo: total",
        ],
    ]
    working_age = pd.read_excel(
        sources["population"], skiprows=7, index_col=0, nrows=92
    ).dropna(how="all")
    rates.columns = rates.columns.str.replace(": total", "")

    ages = list(range(14, 90)) + ["90 y más"]
    working_age = working_age.loc[ages].sum()
    working_age.index = pd.date_range(
        start="1996-06-30", end="2050-06-30", freq="YE-JUN"
    )
    monthly_working_age = working_age.resample("ME").interpolate("linear")
    monthly_working_age = monthly_working_age.reindex(rates.index)
    persons = rates.iloc[:, [0, 1]].div(100).mul(monthly_working_age, axis=0)
    persons["Desempleados"] = rates.iloc[:, 2].div(100).mul(persons.iloc[:, 0])
    persons.columns = ["Activos", "Empleados", "Desempleados"]

    output = pd.concat([rates, persons], axis=1)
    output = output.rename_axis(None)

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Labor market",
        "currency": None,
        "inflation_adjustment": None,
        "unit": "Rate",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    for indicator, unit in zip(ids[-3:], ["Persons", "Persons", "Persons"]):
        metadata.update_indicator_metadata_value(indicator, "unit", unit)
    dataset = Dataset(name, output, metadata)

    return dataset


def real_wages(*args, **kwargs) -> Dataset:
    """
    Get real wages.

    Returns
    -------
    Real wages data : Dataset

    """
    name = get_name_from_function()

    nominal_wages = load_dataset("nominal_wages", *args, **kwargs)
    output = nominal_wages.convert("real").rebase("2008-07-31").to_named()
    output.columns = [
        "Índice medio de salarios reales",
        "Índice medio de salarios reales privados",
        "Índice medio de salarios reales públicos",
    ]

    spanish_names = output.columns
    spanish_names = [{"es": x} for x in spanish_names]
    ids = [f"{name}_{i}" for i in range(output.shape[1])]
    output.columns = ids

    base_metadata = {
        "area": "Labor market",
        "currency": "UYU",
        "inflation_adjustment": "Const.",
        "unit": "2008-07=100",
        "seasonal_adjustment": None,
        "frequency": "ME",
        "time_series_type": "Stock",
        "cumulative_periods": 1,
        "transformations": [],
    }
    metadata = DatasetMetadata.from_cast(
        name, base_metadata, output.columns, spanish_names
    )
    dataset = Dataset(name, output, metadata)

    return dataset
